"""
Utilities for copying metadata to new NeXus files.
"""

from __future__ import annotations

from typing import Any, Dict, List, Tuple

import h5py
import numpy as np
from numpy.typing import ArrayLike

from ..beamlines.beamline_utils import PumpProbe
from ..beamlines.SSX_chip import Chip, compute_goniometer
from ..nxs_utils import Axis, TransformationType
from ..nxs_utils.ScanUtils import calculate_scan_points
from ..nxs_write.write_utils import create_attributes
from ..utils import units_of_length, walk_nxs


def h5str(h5_value: str | np.string_ | bytes) -> str:
    """
    Convert a value returned an h5py attribute to str.

    h5py can return either a bytes-like (numpy.string_) or str object
    for attribute values depending on whether the value was written as
    fixed or variable length. This function collapses the two to str.

    Args:
        h5_value (str | np.string_ | bytes): Original attribute value.

    Returns:
        str: Attribute value collapsed to str.
    """
    if isinstance(h5_value, (np.string_, bytes)):
        return h5_value.decode("utf-8")
    return h5_value


def get_skip_list(nxentry: h5py.Group, skip_obj: List[str]) -> List[str]:
    """
    Get a list of all the objects that should not be copied in the new NeXus file.

    Args:
        nxentry (h5py.Group): NXentry group of a NeXus file.
        skip_obj (List[str]): List of objects that should not be copied.

    Returns:
        skip_list (List[str]): List of "NXclass" objects to skip during copy.
    """
    obj_list = walk_nxs(nxentry)
    skip_list = []
    for obj in obj_list:
        try:
            if nxentry[obj].attrs["NX_class"] in np.string_(skip_obj):
                skip_list.append(obj)
        except Exception:
            pass
    return skip_list


def get_nexus_tree(
    nxs_in: h5py.File,
    nxs_out: h5py.File,
    skip: bool = True,
    skip_obj: List[str] = None,
) -> h5py.Group | None:
    """
    Copy the tree from the original NeXus file. Everything except NXdata is copied to a new NeXus file.
    If skip is False, then the full tree is copied.

    Args:
        nxs_in (h5py.File): Original NeXus file.
        nxs_out (h5py.File): New NeXus file.
        skip (bool, optional): Copy everything but objects in skip_obj, which always include NXdata.
                            Pass False to copy the whole NXentry tree. Defaults to True.
        skip_obj (List[str], optional): List of NX_class objects not to be copied, eg. 'NXdata' or 'NXdetector'.. Defaults to None.

    Returns:
        h5py.Group | None: The group NXentry or nothing if the full file is copied.
    """
    skip_obj = ["NXdata"] if skip_obj is None else skip_obj

    if skip is True:
        nxentry = nxs_out.create_group("entry")
        create_attributes(nxentry, ("NX_class", "default"), ("NXentry", "data"))
        # Copy all of the nexus tree as it is except for the group passed as skip_obj
        skip_list = get_skip_list(nxs_in["entry"], skip_obj)
        # First copy full nxentry
        for k in nxs_in["entry"].keys():
            nxs_in["entry"].copy(k, nxentry)
        # Then delete objects from skip list
        for s in skip_list:
            del nxentry[s]
        return nxentry
    else:
        # Copy everything
        nxs_in.copy("entry", nxs_out)
        return None


def identify_tristan_scan_axis(nxs_in: h5py.File) -> Tuple[str | None, Dict[str, Any]]:
    """
    Identify the scan_axis in the NeXus tree of a Tristan collection.

    Return the first data set in the group '/entry/data' that has the attribute
    'transformation_type' equal to 'rotation'.

    Args:
        nxs_in (h5py.File): Tristan NeXus file

    Returns:
        ax (str | None): Name of the scan_axis.
        ax_attrs (Dict[str, Any]): Attributes of the scan_axis dataset.
    """
    nxdata = nxs_in["entry/data"]
    for ax, h5_object in nxdata.items():
        if h5str(h5_object.attrs.get("transformation_type")) == "rotation":
            return ax, dict(h5_object.attrs)
    return None, {}


def convert_scan_axis(
    nxsample: h5py.Group, nxdata: h5py.Group, ax: str, ax_range: ArrayLike | None = None
):
    """
    Modify all instances of scan_axis present in NeXus file NXsample group.

    Args:
        nxsample (h5py.Group): NXsample group of NeXus file to be modified.
        nxdata (h5py.Group): NXdata group of NeXus file to be modified.
        ax (str): Name of scan_axis.
        ax_range (ArrayLike): Scan points. If passed, axis_increment_set and axis_end will also be written.\
            Defaults to None
    """
    del nxsample["transformations/" + ax]
    nxsample["transformations/" + ax] = nxdata[ax]
    name = (
        "sample_" + ax + "/" + ax if "sam" not in ax else "sample_" + ax[-1] + "/" + ax
    )
    del nxsample[name]
    nxsample[name] = nxdata[ax]
    if ax_range is not None and "sam" not in ax:
        increment = round(ax_range[1] - ax_range[0], 3)
        nxsample["sample_" + ax].create_dataset(ax + "_increment_set", data=increment)
        nxsample["sample_" + ax].create_dataset(ax + "_end", data=ax_range + increment)


def check_and_fix_det_axis(nxs_in: h5py.File):
    det_z_grp = nxs_in["/entry/instrument/detector/transformations/detector_z"]
    det_z = det_z_grp["det_z"]
    if isinstance(det_z[()], bytes) or isinstance(det_z[()], str):
        det_z_attrs = {}
        for k, v in det_z.attrs.items():
            det_z_attrs[k] = v
        dist = float(det_z[()])
        dist = units_of_length(str(dist) + "mm")

        del nxs_in["/entry/instrument/detector/transformations/detector_z/det_z"]
        nxdet_z = det_z_grp.create_dataset("det_z", data=np.array([dist.magnitude]))
        for k, v in det_z_attrs.items():
            nxdet_z.attrs.create(k, v)

        del nxs_in["/entry/instrument/detector/distance"]
        nxs_in["/entry/instrument/detector"].create_dataset(
            "distance", data=dist.to("m").magnitude
        )
        nxs_in["/entry/instrument/detector/distance"].attrs.create(
            "units", np.string_("m")
        )
    else:
        return


def is_chipmap_in_tristan_nxs(
    nxobj: h5py.File | h5py.Group, loc: str = "entry/source/notes/chipmap"
) -> bool:
    """
    Look for the saved chipmap for a SSX experiment inside a tristan nexus file.

    Args:
        nxobj (h5py.File | h5py.Group): NeXus object to be searched, could be a file or a group.
        loc (str, optional): Location where the chipmap should be saved. Defaults to "entry/source/notes/chipmap".

    Returns:
        bool: Returns True is a chipmap is found, False otherwise.
    """
    obj_list = walk_nxs(nxobj)
    if loc in obj_list:
        return True
    else:
        return False


def compute_ssx_axes(
    nxs_in: h5py.File, nbins: int, rot_ax: str, rot_val: Tuple | List | ArrayLike
) -> Tuple[Dict, Dict, Dict, int | None]:
    """
    Computes the positions on chip corresponding to the binned images from a Tristan fixed target collection.

    The function looks for the blocks (chipmap) and the chip_info dictionaries inside the original NeXus file and calculates the scan points from there.
    For older versions of the SSX NeXus files, this information is not available and the scan points will be calculated based on the number of images and
    using the default Oxford chip dimensions, starting from the upper left corned of the chip.
    If multiple windows have been binned into a single image, instead of the sam_(x,y) translation axes values, the number of windows per images will
    be returned and saved in the NeXus file.

    Args:
        nxs_in (h5py.File): File handle for the original Tristan collection NeXus file.
        nbins (int): Number of images.
        rot_ax (str): Rotation axis.
        rot_val (Tuple | List | ArrayLike): Rotation axis start and stop values, as found in the original NeXus.

    Returns:
        OSC, TRANSL, pump_info, windows_per_bin (Tuple[Dict, Dict, Dict, int | None]): Oscillation range, Translation range, pump_probe info, number of windows per binned image.
    """
    # Define pump_probe
    pp = PumpProbe()
    pp.status = True

    # Get chipmap: use default chipmap location: /entry/source/notes/chipmap if there
    if is_chipmap_in_tristan_nxs(nxs_in) is True:
        blocks = eval(nxs_in["/entry/source/notes/chipmap"][()])
        chip_info = eval(nxs_in["/entry/source/notes/chip"][()])
        pp.exposure = chip_info["LASER_DWELL"]
        pp.delay = chip_info["LASER_DELAY"]
    else:
        # Older version of these files don't have any chip information inside.
        from ..beamlines.SSX_chip import CHIP_DICT_DEFAULT

        chip_info = {k: v[1] for k, v in CHIP_DICT_DEFAULT.items()}
        # Assume 400 imgs each block in the chip
        n_blocks = nbins // 400
        key = [f"%0{2}d" % bl for bl in range(1, n_blocks + 1)]
        val = []
        for i in range(chip_info["X_NUM_BLOCKS"]):
            for j in range(chip_info["Y_NUM_BLOCKS"]):
                val.append((i, j))
        blocks = {key[i]: val[i] for i in range(len(key))}
        pp.exposure = nxs_in["/entry/source/notes/pump_exposure_time"][()]
        pp.delay = nxs_in["/entry/source/notes/pump_delay"][()]

    # Define chip
    chip = Chip(
        "fastchip",
        num_steps=[chip_info["X_NUM_STEPS"], chip_info["Y_NUM_STEPS"]],
        step_size=[chip_info["X_STEP_SIZE"], chip_info["Y_STEP_SIZE"]],
        num_blocks=[chip_info["X_NUM_BLOCKS"], chip_info["Y_NUM_BLOCKS"]],
        block_size=[chip_info["X_BLOCK_SIZE"], chip_info["Y_BLOCK_SIZE"]],
        start_pos=[
            chip_info["X_START"],
            chip_info["Y_START"],
            chip_info["Z_START"],
        ],
    )

    # Rotation values
    osc_ax = Axis(
        rot_ax,
        "",
        TransformationType.ROTATION,
        (0, 0, 0),
        start_pos=rot_val[0],
    )
    OSC = calculate_scan_points(osc_ax, rotation=True, tot_num_imgs=nbins)

    num_blocks = (
        chip.tot_blocks() if list(blocks.values())[0] == "fullchip" else len(blocks)
    )

    # Calculate scan start/end positions on chip
    N_EXP = chip_info["N_EXPOSURES"]
    if nbins % (num_blocks * chip.tot_windows_per_block()) == 0:
        # All the windows in the selected blocks (full chip or not) have been scanned at least once.
        N_EXP = nbins // (num_blocks * chip.tot_windows_per_block())

        axis1 = Axis("sam_y", "", TransformationType.TRANSLATION, (0, 0, 0))
        axis2 = Axis("sam_x", "", TransformationType.TRANSLATION, (0, 0, 0))
        start_pos = compute_goniometer(chip, blocks=blocks, ax1="sam_y", ax2="sam_x")

        # Translation values
        TRANSL = {"sam_y": np.array([]), "sam_x": np.array([])}
        for k, v in start_pos.items():
            axis1.start_pos = v[axis1.name]
            axis1.increment = chip.step_size[1] * v["direction"]
            axis1.num_steps = chip.num_steps[1]
            axis2.start_pos = v[axis2.name]
            axis2.increment = chip.step_size[0]
            axis2.num_steps = chip.num_steps[0]

            transl = calculate_scan_points(axis1, axis2)

            TRANSL["sam_y"] = np.append(TRANSL["sam_y"], np.round(transl["sam_y"], 3))
            TRANSL["sam_x"] = np.append(TRANSL["sam_x"], np.round(transl["sam_x"], 3))

        if N_EXP > 1:
            OSC = {k: [val for val in v for _ in range(N_EXP)] for k, v in OSC.items()}
            TRANSL = {
                k: [val for val in v for _ in range(N_EXP)] for k, v in TRANSL.items()
            }

        # Update pump probe
        pump_info = pp.to_dict()
        pump_info["n_exposures"] = N_EXP
        return OSC, TRANSL, pump_info, None
    else:
        # Hopefully this will not happen..
        # Do not set N_EXP, there should be no repeat here
        # Find out how many consecutive windows are binned together.
        windows_per_bin = (num_blocks * chip.tot_windows_per_block()) // nbins
        return OSC, {}, pp.to_dict(), windows_per_bin



"""
Utilities to look for scan axes and calculate scan ranges from a list of Axis objects.
"""

from __future__ import annotations

import logging
from collections import namedtuple
from enum import IntEnum
from typing import Dict, List

from numpy.typing import ArrayLike
from scanspec.core import Path as ScanPath
from scanspec.specs import Line

from .. import log
from .Axes import Axis

scan_logger = logging.getLogger("nexgen.ScanUtils")
log.config()

# Some options for grid scans
GridScanOptions = namedtuple("GridScanOptions", ("axes_order", "snaked"))
GridScanOptions.__doc__ = "Options for defining a grid scan"
GridScanOptions.axes_order.__doc__ = "List of axes in order of (fast, slow)."
GridScanOptions.snaked.__doc__ = "Boolean to say whether it's a snaked scan."


class ScanDirection(IntEnum):
    POSITIVE = 1
    NEGATIVE = -1


class ScanAxisNotFoundError(Exception):
    def __init__(self, errmsg):
        scan_logger.error(errmsg)


class ScanAxisError(Exception):
    def __init__(self, errmsg):
        scan_logger.error(errmsg)


def identify_osc_axis(
    axes_list: List[Axis],
    default: str = "omega",
) -> str:
    """
    Identify the rotation scan_axis.

    This function identifies the scan axis from the list passed as argument.
    The scan axis is the one where start and end value are not the same.
    If there is only one rotation axis, that is the one returned.
    In the case scan axis cannot be identified, a default value is arbitrarily assigned.

    Args:
        axes_list (List[Axis]): List of axes objects associated to goniometer axes.
        default (str, optional): String to deafult to in case scan axis is not found. Defaults to "omega".

    Raises:
        ScanAxisNotFoundError: If no axes have been passed.
        ValueError: If more than one rotation axis seems to move.

    Returns:
        scan_axis (str): String identifying the rotation scan axis.
    """
    # This assumes that at least one axis is always passed.
    if len(axes_list) == 0:
        raise ScanAxisNotFoundError(
            "Impossible to determine oscillation scan axis. No axes passed to function. Please make sure at least one value is passed."
        )
    # Look only for rotation axes
    rot_axes = [ax for ax in axes_list if ax.transformation_type == "rotation"]
    if len(rot_axes) == 1:
        return rot_axes[0].name
    else:
        scan_idx = [ax.is_scan for ax in rot_axes]
        if scan_idx.count(True) == 0:
            return default
        if scan_idx.count(True) > 1:
            raise ScanAxisNotFoundError(
                "Unable to correctly identify the rotation scan axis."
            )
        return rot_axes[scan_idx.index(True)].name


def identify_grid_scan_axes(
    axes_list: List[Axis],
) -> List[str]:
    """
    Identify the scan axes for a translation linear/grid scan.

    Args:
        axes_list (List[Axis]): List of axes objects associated to goniometer axes.

    Raises:
        ScanAxisNotFoundError: If no axes have been passed.

    Returns:
        scan_axis (List[str]): List of strings identifying the linear/grid scan axes. If no axes are identified, it will return an empty list.
    """
    if len(axes_list) == 0:
        raise ScanAxisNotFoundError(
            "Impossible to determine translation scan. No axes passed to function. Please make sure at least one value is passed."
        )

    # Look only at translation axes
    tr_axes = [ax for ax in axes_list if ax.transformation_type == "translation"]
    grid_axes = [ax.name for ax in tr_axes if ax.is_scan is True]
    return grid_axes


def calculate_scan_points(
    axis1: Axis,
    axis2: Axis | None = None,
    snaked: bool = True,
    rotation: bool = False,
    tot_num_imgs: int | None = None,
) -> Dict[str, ArrayLike]:
    """Calculate the scan range for a linear/grid scan or a rotation scan from the number of images (steps) to be written in each direction.

    When dealing with a rotation axis, if there are multiple images but no rotation scan, return the axis start position repeated as many times \
        as the number of images - either defined by the num_steps attribute of the Axis object or passed as tot_num_imgs.

    Args:
        axis1 (Axis): Axis object describing the axis involved in a scan.
        axis2 (Axis, optional): Axis object describing the second axis involved in a scan. Only necessary for a grid scan. Defaults to None.
        snaked (bool, optional):  If True, scanspec will "draw" a grid where the second axis is snaked. \
            It will be ignored for a rotation scan. Defaults to True.
        rotation (bool, optional): Tell the function to calculate a rotation scan. Defaults to False.
        tot_num_imgs (int, optional): Total number of images. Only used for oscillation axis when there is no rotation. \
            It will be ignored otherwise. Defaults to None.

    Raises:
        ScanAxisError: If the passed axis has the wrong transformation type.
        ValueError: For a rotation axis with no rotation, if the number of images is missing.

    Returns:
        Dict[str, ArrayLike]: A dictionary of ("axis_name": axis_range) key-value pairs.
    """

    if rotation is True:
        if axis1.transformation_type != "rotation":
            raise ScanAxisError(
                f"Wrong transformation type: a {axis1.transformation_type} has been passed for a rotation scan."
            )
        if axis1.num_steps == 0 and tot_num_imgs is None:
            raise ValueError(
                "Missing number of scan points, impossible to calculate scan."
            )

        n_images = tot_num_imgs if tot_num_imgs else axis1.num_steps
        spec = Line(axis1.name, axis1.start_pos, axis1.end_pos, n_images)
        scan_path = ScanPath(spec.calculate())

        return scan_path.consume().midpoints

    if axis1.transformation_type != "translation":
        raise ScanAxisError(
            f"Wrong transformation type: a {axis1.transformation_type} has been passed for a translation scan."
        )

    if axis2 and axis2.transformation_type != "translation":
        raise ScanAxisError(
            f"Wrong transformation type: a {axis2.transformation_type} has been passed for a translation scan."
        )

    if axis2 is None:
        spec = Line(
            axis1.name,
            axis1.start_pos,
            axis1.end_pos,
            axis1.num_steps,
        )
    elif axis2 and snaked is True:
        spec = Line(
            axis1.name,
            axis1.start_pos,
            axis1.end_pos,
            axis1.num_steps,
        ) * ~Line(
            axis2.name,
            axis2.start_pos,
            axis2.end_pos,
            axis2.num_steps,
        )
    else:
        spec = Line(
            axis1.name,
            axis1.start_pos,
            axis1.end_pos,
            axis1.num_steps,
        ) * Line(
            axis2.name,
            axis2.start_pos,
            axis2.end_pos,
            axis2.num_steps,
        )
    scan_path = ScanPath(spec.calculate())

    return scan_path.consume().midpoints



from __future__ import annotations

from pathlib import Path
from unittest.mock import patch

import pytest

from nexgen.beamlines.I19_2_nxs import ExperimentTypeError, eiger_writer, nexus_writer


def test_nexus_writer_raises_error_for_serial():
    with pytest.raises(ExperimentTypeError):
        nexus_writer("/path/to/meta", "eiger", 0.1, serial=True)


@patch("nexgen.beamlines.I19_2_nxs.log")
def test_nexus_writer_fails_if_missing_n_imgs_and_not_using_meta(fake_log):
    with pytest.raises(ValueError):
        nexus_writer("/path/to/meta", "eiger", 0.1, use_meta=False)


@patch("nexgen.beamlines.I19_2_nxs.CollectionParams")
def test_eiger_nxs_writer_fails_if_missing_axes_and_no_meta(fake_params):
    with pytest.raises(ValueError):
        eiger_writer(Path("/path/to/meta"), fake_params, use_meta=False)



import logging
from unittest.mock import MagicMock

import pytest

from nexgen import log


@pytest.fixture
def dummy_logger():
    logger = logging.getLogger("nexgen")
    yield logger


def test_basic_logging_config(dummy_logger):
    assert dummy_logger.hasHandlers() is True
    assert len(dummy_logger.handlers) == 1
    assert dummy_logger.handlers[0].level == logging.DEBUG


def test_logging_config_with_filehandler(dummy_logger):
    logfile = MagicMock()
    log.config(logfile, delayed=True)
    assert len(dummy_logger.handlers) == 2
    assert dummy_logger.handlers[1].level == logging.DEBUG
    # Clear FileHandler to avoid other tests failing if open
    dummy_logger.removeHandler(dummy_logger.handlers[1])



import time
from pathlib import Path

import numpy as np
import pint
import pytest

from nexgen import utils

ureg = pint.UnitRegistry()


def test_cif2nxs():
    assert utils.imgcif2mcstas([0, 0, 0]) == (0, 0, 0)
    assert utils.imgcif2mcstas([1, 0, 0]) == (-1, 0, 0)
    assert utils.imgcif2mcstas([0, 1, 0]) == (0, 1, 0)
    assert utils.imgcif2mcstas([0, 0, 1]) == (0, 0, -1)


def test_coord2nxs():
    R = np.array([[1, 0, 0], [0, 0, -1], [0, 1, 0]])
    assert utils.coord2mcstas([0, 0, 0], R) == (0, 0, 0)
    assert utils.coord2mcstas([1, 0, 0], R) == (1, 0, 0)
    assert utils.coord2mcstas([0, 1, 0], R) == (0, 0, 1)
    assert utils.coord2mcstas([0, 0, 1], R) == (0, -1, 0)


def test_coerce_to_path():
    s = "/some/path"
    s = utils.coerce_to_path(s)
    assert isinstance(s, Path)


def test_find_in_dict():
    test_dict = {
        "alpha": 0,
        "beta": 1,
    }
    assert utils.find_in_dict("alpha", test_dict) is True
    assert utils.find_in_dict("gamma", test_dict) is False


def test_get_filename_template():
    # Check filename from _master.h5 file
    fn = utils.get_filename_template(Path("File_01_master.h5"))
    assert isinstance(fn, str)
    assert fn == "File_01_%06d.h5"
    assert fn % 1 == "File_01_000001.h5"
    # Check filename from .nxs file
    fn = utils.get_filename_template(Path("File_02.nxs"))
    assert isinstance(fn, str)
    assert fn == "File_02_%06d.h5"
    assert fn % 1 == "File_02_000001.h5"


def test_get_nexus_filename():
    # Check nexus filename from meta
    nxs = utils.get_nexus_filename(Path("File_01_meta.h5"))
    assert nxs.as_posix() == "File_01.nxs"
    # Check nexus filename from datafile
    nxs = utils.get_nexus_filename(Path("File_02_0001.h5"))
    assert nxs.as_posix() == "File_02.nxs"


def test_iso_timestamps():
    assert utils.get_iso_timestamp(None) is None
    # Check that no exceptions are raised when passing a time.time() object
    assert utils.get_iso_timestamp(time.time())


def test_units_of_length():
    assert utils.units_of_length("1.5m") == ureg.Quantity(1.5, "m")
    # Check that a dimensionless unit defaults to mm
    assert utils.units_of_length(100) == ureg.Quantity(100, "m")
    # Check conversion to base units
    assert utils.units_of_length("5cm", True) == ureg.Quantity(0.05, "m")
    assert utils.units_of_length("1in", True) == ureg.Quantity(0.0254, "m")


def test_units_of_length_raises_error_for_wrong_dimension():
    with pytest.raises(pint.errors.DimensionalityError):
        utils.units_of_length("30s")


def test_units_of_time():
    assert utils.units_of_time("0.05s") == ureg.Quantity(0.05, "s")
    # Check that a dimensionless value deafults to seconds
    assert utils.units_of_time(1) == ureg.Quantity(1, "s")
    # Check conversion to base units
    assert utils.units_of_time("20ms") == ureg.Quantity(0.02, "s")


def test_units_of_time_raises_error_for_wrong_dimension():
    with pytest.raises(pint.errors.DimensionalityError):
        utils.units_of_time("0.2in")



"""
Define beamline parameters for I19-2, Tristan and Eiger detectors.
"""

from ..nxs_utils import Axis, TransformationType
from ..utils import Point3D
from .beamline_utils import BeamlineAxes

I19_2_gonio = [
    Axis("omega", ".", TransformationType.ROTATION, Point3D(-1, 0, 0)),
    Axis(
        "kappa", "omega", TransformationType.ROTATION, Point3D(-0.642788, -0.766044, 0)
    ),
    Axis("phi", "kappa", TransformationType.ROTATION, Point3D(-1, 0, 0)),
    Axis("sam_z", "phi", TransformationType.TRANSLATION, Point3D(0, 0, 1)),
    Axis("sam_y", "sam_z", TransformationType.TRANSLATION, Point3D(0, 1, 0)),
    Axis("sam_x", "sam_y", TransformationType.TRANSLATION, Point3D(1, 0, 0)),
]

I19_2Eiger = BeamlineAxes(
    gonio=I19_2_gonio,
    det_axes=[
        Axis("two_theta", ".", TransformationType.ROTATION, Point3D(-1, 0, 0)),
        Axis("det_z", "two_theta", TransformationType.TRANSLATION, Point3D(0, 0, 1)),
    ],
    fast_axis=Point3D(0, 1, 0),
    slow_axis=Point3D(-1, 0, 0),
)

I19_2Tristan = BeamlineAxes(
    gonio=I19_2_gonio,
    det_axes=[
        Axis("two_theta", ".", TransformationType.ROTATION, Point3D(-1, 0, 0), 0),
        Axis("det_z", "two_theta", TransformationType.TRANSLATION, Point3D(0, 0, 1)),
    ],
    fast_axis=Point3D(-1, 0, 0),
    slow_axis=Point3D(0, 1, 0),
)



# import pytest
from numpy.testing import assert_array_equal

from nexgen.nxs_utils import (
    Axis,
    Detector,
    DetectorModule,
    EigerDetector,
    JungfrauDetector,
    SinglaDetector,
    TransformationType,
    TristanDetector,
)
from nexgen.utils import Point3D

det_axes = [
    Axis("two_theta", ".", TransformationType.ROTATION, Point3D(1, 0, 0), 0.0),
    Axis("det_z", "two_theta", TransformationType.TRANSLATION, Point3D(0, 0, 1), 500.0),
]

test_eiger = EigerDetector("Eiger2 1M", (1028, 1062), "Si", 10000, -1)
test_tristan = TristanDetector("Tristan 1M", (515, 2069))
test_jungfrau = JungfrauDetector("Jungfrau 1M", (1066, 1030))
test_singla = SinglaDetector("Singla 1M", (1062, 1028))


def test_eiger_detector():
    assert test_eiger.description == "Eiger2 1M"
    assert test_eiger.sensor_material == "Si"
    assert test_eiger.sensor_thickness == "0.450mm"
    assert test_eiger.pixel_size == ["0.075mm", "0.075mm"]
    assert test_eiger.hasMeta is True


def test_tristan_detector():
    assert test_tristan.description == "Tristan 1M"
    assert test_tristan.sensor_material == "Si"
    assert test_tristan.sensor_thickness == "0.5mm"
    assert test_tristan.mode == "events"
    assert test_eiger.hasMeta is True


def test_jungfrau_detector():
    assert test_jungfrau.description == "Jungfrau 1M"
    assert test_jungfrau.sensor_material == "Si"
    assert test_jungfrau.sensor_thickness == "0.320mm"
    assert test_jungfrau.hasMeta is False
    assert isinstance(test_jungfrau.constants, dict)


def test_singla_detector():
    assert test_singla.description == "Singla 1M"
    assert test_singla.sensor_thickness == "0.450mm"
    assert test_singla.detector_type == "HPC"
    assert test_singla.hasMeta is False


def test_detector_axes():
    det = Detector(
        test_eiger, det_axes, [100, 200], 0.1, [(0, 0, 1), Point3D(0, -1, 0)]
    )
    assert isinstance(det.detector_axes, list)
    assert det.detector_axes[0].name == "two_theta"
    assert det.detector_axes[1].name == "det_z"
    assert [det.detector_axes[0].depends, det.detector_axes[1].depends] == [
        ".",
        "two_theta",
    ]


def test_get_detector_description():
    eig = Detector(
        test_eiger, det_axes, [100, 200], 0.1, [(0, 0, 1), Point3D(0, -1, 0)]
    )
    assert eig.get_detector_description() == test_eiger.description


def test_get_detector_mode():
    eig = Detector(
        test_eiger, det_axes, [100, 200], 0.1, [(0, 0, 1), Point3D(0, -1, 0)]
    )
    assert eig.get_detector_mode() == "images"

    tr = Detector(
        test_tristan, det_axes, [100, 200], 0.1, [(0, 0, 1), Point3D(0, -1, 0)]
    )
    assert tr.get_detector_mode() == "events"


def test_detector_to_module_dict():
    mod = Detector(
        test_eiger, det_axes, [100, 200], 0.1, [(0, 0, 1), Point3D(0, -1, 0)]
    ).get_module_info()
    assert isinstance(mod, dict)
    assert mod["module_offset"] == "1"
    assert_array_equal(mod["fast_axis"], (0, 0, 1))
    assert_array_equal(mod["slow_axis"], (0, -1, 0))


def test_fast_slow_axis_input():
    mod = DetectorModule((0, 0, 1), Point3D(0, -1, 0))

    assert isinstance(mod.fast_axis, tuple) and isinstance(mod.slow_axis, tuple)
    assert mod.fast_axis == (0, 0, 1)
    assert mod.slow_axis == (0, -1, 0)
    assert mod.module_offset == "1"



"""General utilities for comamnd line tools"""

import argparse
from pathlib import Path

from .. import __version__

version_parser = argparse.ArgumentParser(add_help=False)
version_parser.add_argument(
    "-V",
    "--version",
    action="version",
    version=f"%(prog)s: NeXus generation tools {__version__}",
)

config_parser = argparse.ArgumentParser(add_help=False)
config_parser.add_argument(
    "-c",
    "--show-config",
    action="store_true",
    default=False,
    dest="show_config",
    help="Show the configuration parameters.",
)
config_parser.add_argument(
    "-a",
    "--attributes-level",
    default=0,
    type=int,
    dest="attributes_level",
    help="Set the attributes level for showing the configuration parameters.",
)

detectormode_parser = argparse.ArgumentParser(add_help=False)
group = detectormode_parser.add_mutually_exclusive_group(required=True)
group.add_argument(
    "-i",
    "--images",
    help="Write a demo file with blank images.",
    action="store_true",
)
group.add_argument(
    "-e",
    "--events",
    help="Write a demo file with fake events.",
    action="store_true",
)
detectormode_parser.add_argument(
    "-f",
    "--force",
    help="Overrides other instructions relevant to number of images/ "
    "stream of events already parsed. For images, please pass the desired number to be written. "
    "For events, the number of chunks to be written per file. "
    "The number of files will be determined by the number of detector modules.",
    type=int,
    default=None,
)


class _CheckFileExtension(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        condition = any("filename" in v for v in values)
        if condition is True:
            i = ["filename" in v for v in values].index(True)
            fname = Path(values[i]).expanduser().resolve()
            ext = fname.suffix
            if ext != ".h5" and ext != ".nxs":
                print(
                    f"You specified an invalid extension {ext} for the output file.\n"
                    f"It will be saved to {fname.stem}.nxs instead."
                )
                values[i] = f"{fname.stem}.nxs"
        setattr(namespace, self.dest, values)


# Define subparsers for NeXus generator
nexus_parser = argparse.ArgumentParser(add_help=False)
nexus_parser.add_argument("phil_args", nargs="*")

demo_parser = argparse.ArgumentParser(add_help=False)
demo_parser.add_argument("phil_args", nargs="*", action=_CheckFileExtension)

# Define subparsesrs for NeXus copy
full_copy_parser = argparse.ArgumentParser(add_help=False)
full_copy_parser.add_argument("phil_args", nargs="*")

tristan_copy_parser = argparse.ArgumentParser(add_help=False)
tristan_copy_parser.add_argument("phil_args", nargs="*")
tristan_group = tristan_copy_parser.add_mutually_exclusive_group(
    required=False
)  # required=True)
tristan_group.add_argument(
    "-o",
    "--osc-angle",
    help="Oscillation angle, in degrees",
    type=float,
)
tristan_group.add_argument(
    "-n",
    "--num-bins",
    help="Number of binnes images",
    type=int,
)



"""
Define parameters for Electron Diffraction with Singla detector.
"""

from ..nxs_utils import Axis, Facility, Source, TransformationType
from ..utils import Point3D
from .beamline_utils import BeamlineAxes

EDSingla = BeamlineAxes(
    gonio=[
        Axis("alpha", ".", TransformationType.ROTATION, Point3D(-1, 0, 0)),
        Axis("sam_z", "alpha", TransformationType.TRANSLATION, Point3D(0, 0, 1)),
        Axis("sam_y", "sam_z", TransformationType.TRANSLATION, Point3D(0, 1, 0)),
        Axis("sam_x", "sam_y", TransformationType.TRANSLATION, Point3D(1, 0, 0)),
    ],
    det_axes=[Axis("det_z", ".", TransformationType.TRANSLATION, Point3D(0, 0, 1))],
    fast_axis=Point3D(-1, 0, 0),
    slow_axis=Point3D(0, -1, 0),
)

coordinate_frame = "mcstas"

ED_coord_system = {
    "convention": "ED",
    "origin": (0, 0, 0),
    "x": Axis("x", ".", TransformationType.TRANSLATION, Point3D(0, 1, 0)),
    "y": Axis("y", "x", TransformationType.TRANSLATION, Point3D(-1, 0, 0)),
    "z": Axis("z", "y", TransformationType.TRANSLATION, Point3D(0, 0, 1)),
}

EDSource = Source(
    "eBIC",
    Facility("Diamond Light Source", "DLS", "Electron Source", "DIAMOND MICROSCOPE"),
    "electron",
)



"""
Object definition for goniometer.
"""

from __future__ import annotations

from typing import Dict, List, Tuple

from numpy.typing import ArrayLike

from .Axes import Axis
from .ScanUtils import (
    GridScanOptions,
    ScanDirection,
    calculate_scan_points,
    identify_grid_scan_axes,
    identify_osc_axis,
)


class Goniometer:
    """
    Goniometer definition.

    Attributes:
        axes_list: List of axes making up the goniometer, including their vectors and positions.
        scan: The scan executed during the collection, could be a rotation or a grid scan. If not passed can be updated from the axes.
    """

    def __init__(
        self,
        axes: List[Axis],
        scan: Dict[str, ArrayLike] | None = None,
    ):
        self.axes_list = axes
        self.scan = scan
        if self.scan:
            self._check_and_update_goniometer_from_scan(list(self.scan.keys()))

    def __repr__(self) -> str:
        msg = ""
        for ax in self.axes_list:
            msg += f"{ax.name}: {ax.start_pos} => {ax.transformation_type} on {ax.depends} \n\t"
        if self.scan:
            msg += f"Scan axis/axes: {list(self.scan.keys())}. \n"
        return f"Goniometer information: \n\t{msg}"

    def _check_and_update_goniometer_from_scan(self, scan_axes: List[str]):
        """Check that the values entered for the goniometer match with the scan."""
        for ax in scan_axes:
            idx = self._find_axis_in_goniometer(ax)
            if self.axes_list[idx].start_pos != self.scan[ax][0]:
                self.axes_list[idx].start_pos = self.scan[ax][0]
            u = self._get_unique_scan_point_values(
                ax
            )  # Re-order them in case of a reverse scan, as unique auto sorts
            if len(u) == 1:
                # eg. for a scan that goes back and forth on one line.
                self.axes_list[idx].increment == 0.0
            if len(u) > 1 and self.axes_list[idx].increment != round(u[1] - u[0], 3):
                self.axes_list[idx].increment = round(u[1] - u[0], 3)
            self.axes_list[idx].num_steps = len(u)

    def _find_axis_in_goniometer(self, val: str) -> int:
        """Find the index of the axis matching the input string."""
        idx = [n for n, ax in enumerate(self.axes_list) if ax.name == val]
        if len(idx) == 0:
            return None
        return idx[0]

    def _get_unique_scan_point_values(self, ax: str) -> List:
        """Get the unique values for a scan, in the order they are collected."""
        # Doing this in place of np.unique which automatically sorts the values, leading to
        # errors in reverse rotation scans. Nothing should change for grid scans.
        scan_array = self.scan[ax]
        val = []
        for i in scan_array:
            if i not in val:
                val.append(i)
        return val

    def define_scan_from_goniometer_axes(
        self,
        grid_scan_options: GridScanOptions | None = None,
        scan_direction: ScanDirection = ScanDirection.POSITIVE,
        update: bool = True,  # Option to set to False for ssx if needed
    ) -> Tuple[Dict, Dict]:
        """Define oscillation and/or grid scan ranges for image data collections."""
        if self.scan:
            # Look at keys to see if rotation or grid scan
            scan_axes = list(self.scan.keys())
            # Actually I just need one
            ax_idx = self._find_axis_in_goniometer(scan_axes[0])
            if self.axes_list[ax_idx].transformation_type == "rotation":
                osc_scan = self.scan
                transl_scan = None

            else:
                # Find number of scan points
                tot_num_imgs = len(self.scan[scan_axes[0]])
                osc_axis = identify_osc_axis(self.axes_list)
                osc_idx = self._find_axis_in_goniometer(osc_axis)
                osc_scan = calculate_scan_points(
                    self.axes_list[osc_idx], rotation=True, tot_num_imgs=tot_num_imgs
                )
                transl_scan = self.scan

            if update is True:
                self._check_and_update_goniometer_from_scan(scan_axes)

            return osc_scan, transl_scan

        osc_axis = identify_osc_axis(self.axes_list)
        osc_idx = self._find_axis_in_goniometer(osc_axis)

        transl_axes = (
            grid_scan_options.axes_order
            if grid_scan_options
            else identify_grid_scan_axes(self.axes_list)
        )

        if len(transl_axes) == 0:
            # Take care of rotations in both directions
            self.axes_list[osc_idx].increment = (
                self.axes_list[osc_idx].increment * scan_direction.value
            )
            osc_scan = calculate_scan_points(self.axes_list[osc_idx], rotation=True)
            return osc_scan, None

        transl_idx = [self._find_axis_in_goniometer(ax) for ax in transl_axes]
        if len(transl_axes) == 1:
            self.axes_list[transl_idx[0]].increment = (
                self.axes_list[transl_idx[0]].increment * scan_direction
            )
            transl_scan = calculate_scan_points(self.axes_list[transl_idx[0]])
        else:
            snaked = True if not grid_scan_options else grid_scan_options.snaked
            transl_scan = calculate_scan_points(
                self.axes_list[transl_idx[0]],
                self.axes_list[transl_idx[1]],
                snaked=snaked,
            )

        tot_num_imgs = len(list(transl_scan.values())[0])
        osc_scan = calculate_scan_points(
            self.axes_list[0], rotation=True, tot_num_imgs=tot_num_imgs
        )

        return osc_scan, transl_scan

    def define_scan_axes_for_event_mode(
        self,
        end_position: float | None = None,
    ) -> Tuple[Dict, Dict]:
        """Define oscillation and/or grid scan ranges for event-mode collections."""
        # NOTE For Tristan we already give it (start, stop).
        # To figure out how this actually will work, I need to fix the Tristan writer.
        if self.scan:
            scan_axis = list(self.scan.keys())
            ax_idx = self._find_axis_in_goniometer(scan_axis[0])
            if self.axes_list[ax_idx].transformation_type == "rotation":
                return self.scan, None
            else:
                # We actually always pass a rotation here but future proofing
                return {"omega": (0.0, 0.0)}, self.scan

        osc_axis = identify_osc_axis(self.axes_list)
        osc_idx = self._find_axis_in_goniometer(osc_axis)
        osc_range = (
            (self.axes_list[osc_idx].start_pos, end_position)
            if end_position
            else (self.axes_list[osc_idx].start_pos, self.axes_list[osc_idx].end_pos)
        )

        return {osc_axis: osc_range}, None

    def get_number_of_scan_points(self):
        """Get the number of scan points from the defined scan."""
        scan = (
            self.scan
            if self.scan is not None
            else self.define_scan_from_goniometer_axes()[0]
        )

        axis_name = list(scan.keys())[0]
        scan_length = len(scan[axis_name])
        return scan_length



"""
A writer for NXmx format NeXus Files.
"""
from __future__ import annotations

import logging
import math
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Tuple

import h5py
import numpy as np
from numpy.typing import DTypeLike

from ..nxs_utils.Detector import Detector
from ..nxs_utils.Goniometer import Goniometer
from ..nxs_utils.Source import Attenuator, Beam, Source
from ..tools.VDS_tools import (
    clean_unused_links,
    image_vds_writer,
    jungfrau_vds_writer,
    vds_file_writer,
)
from ..utils import (
    MAX_FRAMES_PER_DATASET,
    MAX_SUFFIX_DIGITS,
    coord2mcstas,
    get_filename_template,
)
from .NXclassWriters import (
    write_NXcoordinate_system_set,
    write_NXdata,
    write_NXdatetime,
    write_NXdetector,
    write_NXdetector_module,
    write_NXentry,
    write_NXinstrument,
    write_NXnote,
    write_NXsample,
    write_NXsource,
)
from .write_utils import TSdset, calculate_estimated_end_time

# Logger
nxmx_logger = logging.getLogger("nexgen.NXmxFileWriter")
nxmx_logger.setLevel(logging.DEBUG)


class NXmxFileWriter:
    """A class to generate NXmx format NeXus files."""

    def __init__(
        self,
        filename: Path | str,
        goniometer: Goniometer,
        detector: Detector,
        source: Source,
        beam: Beam,
        attenuator: Attenuator,
        tot_num_imgs: int,  # | None = None,
    ):
        self.filename = Path(filename).expanduser().resolve()
        self.goniometer = goniometer
        self.detector = detector
        self.source = source
        self.beam = beam
        self.attenuator = attenuator
        self.tot_num_imgs = tot_num_imgs

    def _get_meta_file(self, image_filename: str = None) -> Path | None:
        """Get filename_meta.h5 file in directory if it's supposed to exist."""
        if self.detector.detector_params.hasMeta is False:
            nxmx_logger.debug("No meta file for this collection.")
            return None
        if image_filename:
            return self.filename.parent / f"{image_filename}_meta.h5"
        else:
            return self.filename.parent / f"{self.filename.stem}_meta.h5"

    def _get_collection_time(self):
        return self.detector.exp_time * self.tot_num_imgs

    def _get_data_files_list(
        self,
        image_filename: str | None = None,
    ) -> List[Path]:
        """Get list of datafiles.

        Args:
            image_filename (str | None, optional): Filename stem to use to look for image files. Needed in case it doesn't match \
                the NeXus file name. Defaults to None.

        Returns:
            List[Path]: List of data files to link to.
        """
        num_files = math.ceil(self.tot_num_imgs / MAX_FRAMES_PER_DATASET)
        template = (
            get_filename_template(self.filename)
            if not image_filename
            else (
                self.filename.parent / f"{image_filename}_%0{MAX_SUFFIX_DIGITS}d.h5"
            ).as_posix()
        )
        datafiles = [Path(template % i) for i in range(1, num_files + 1)]
        nxmx_logger.debug(f"Number of datafiles to be written: {len(datafiles)}.")
        return datafiles

    def update_timestamps(
        self, timestamp: datetime | str, dset_name: TSdset = "end_time"
    ):
        """Save timestamps for start and/or end collection if not written before.

        Args:
            timestamp (datetime | str): Timestamp, as datetime or str.
            dset_name (TSdset, optional): Name of dataset to write to nexus file.\
                Allowed values: ["start_time", "end_time", "end_time_estimated". Defaults to "end_time".
        """
        with h5py.File(self.filename, "r+") as nxs:
            write_NXdatetime(nxs, timestamp, dset_name)
        nxmx_logger.info(f"{dset_name} timestamp for collection updated.")

    def add_NXnote(self, notes: Dict, loc: str = "/entry/notes"):
        """Save any additional information as NXnote at the end of the collection.

        Args:
            notes (Dict): Dictionary of (key, value) pairs where key represents the dataset name and value its data.
            loc (str, optional): Location in the NeXus file to save metadata. Defaults to "/entry/notes".
        """
        with h5py.File(self.filename, "r+") as nxs:
            write_NXnote(nxs, loc, notes)
        nxmx_logger.info(f"Notes saved in {loc}.")

    def write(
        self,
        image_datafiles: List | None = None,
        image_filename: str | None = None,
        start_time: datetime | str | None = None,
        est_end_time: datetime | str | None = None,
        write_mode: str = "x",
    ):
        """Write the NXmx format NeXus file.

        This function calls the writers for the main NXclass objects.

        Args:
            image_datafiles (List | None, optional): List of image data files. If not passed, the program will look for \
                files with the stem_######.h5 in the target directory. Defaults to None.
            image_filename (str | None, optional): Filename stem to use to look for image files. Needed in case it doesn't match \
                the NeXus file name. Format: filename_runnumber. Defaults to None.
            start_time (datetime | str, optional): Collection start time if available, in the format "%Y-%m-%dT%H:%M:%SZ".\
                Defaults to None.
            est_end_time (datetime | str, optional): Collection estimated end time if available, in the format "%Y-%m-%dT%H:%M:%SZ".\
                Defaults to None.
            write_mode (str, optional): String indicating writing mode for the output NeXus file. Accepts any valid \
                h5py file opening mode. Defaults to "x".
        """
        metafile = self._get_meta_file(image_filename)
        if metafile:
            nxmx_logger.debug(f"Metafile name: {metafile.as_posix()}.")

        datafiles = (
            image_datafiles
            if image_datafiles
            else self._get_data_files_list(image_filename)
        )

        module = self.detector.get_module_info()

        osc, transl = self.goniometer.define_scan_from_goniometer_axes()

        with h5py.File(self.filename, write_mode) as nxs:
            # NXentry and NXmx definition
            write_NXentry(nxs)

            # Start and estimated end time if known
            if start_time:
                write_NXdatetime(nxs, start_time, "start_time")
                if not est_end_time:
                    tot_exp_time = self._get_collection_time()
                    est_end = calculate_estimated_end_time(start_time, tot_exp_time)
                    write_NXdatetime(nxs, est_end, "end_time_estimated")
            if est_end_time:
                write_NXdatetime(nxs, est_end_time, "end_time_estimated")

            # NXdata: entry/data
            write_NXdata(
                nxs,
                datafiles,
                self.goniometer.axes_list,
                "images",
                osc,
                transl,
            )

            # NXinstrument: entry/instrument
            write_NXinstrument(
                nxs,
                self.beam,
                self.attenuator,
                self.source,
            )

            # NXdetector: entry/instrument/detector
            write_NXdetector(
                nxs,
                self.detector,
                self.tot_num_imgs,
                metafile,
            )

            # NXmodule: entry/instrument/detector/module
            write_NXdetector_module(
                nxs,
                module,
                self.detector.detector_params.image_size,
                self.detector.detector_params.pixel_size,
                beam_center=self.detector.beam_center,
            )

            # NXsource: entry/source
            write_NXsource(nxs, self.source)

            # NXsample: entry/sample
            write_NXsample(
                nxs,
                self.goniometer.axes_list,
                "images",
                osc,
                transl,
                sample_depends_on=None,  # TODO
            )

    def write_vds(
        self,
        vds_offset: int = 0,
        vds_shape: Tuple[int, int, int] = None,
        vds_dtype: DTypeLike = np.uint16,
        clean_up: bool = False,
    ):
        """Write a Virtual Dataset.

        This method adds a VDS under /entry/data/data in the NeXus file, linking to either the full datasets or the subset defined by \
        vds_offset (used as start index) and vds_shape.
        WARNING. Only use clean up if the data collection is finished and all the files have already been written.

        Args:
            vds_offset (int, optional): Start index for the vds writer. Defaults to 0.
            vds_shape (Tuple[int,int,int], optional): Shape of the data which will be linked in the VDS. If not passed, it will be defined as \
            (tot_num_imgs - start_idx, *image_size). Defaults to None.
            vds_dtype (DTypeLike, optional): The type of the input data. Defaults to np.uint16.
            clean_up(bool, optional): Clean up unused links in vds. Defaults to False.
        """
        if not vds_shape:
            vds_shape = (
                self.tot_num_imgs - vds_offset,
                *self.detector.detector_params.image_size,
            )
            if self.goniometer.get_number_of_scan_points() != vds_shape[0]:
                vds_shape = (
                    self.goniometer.get_number_of_scan_points(),
                    *vds_shape[1:],
                )
                nxmx_logger.warning(
                    "The number of scan points doesn't match the calculated vds_shape. \
                    Resetting it to match the number of frames indicated by the scan."
                )

        nxmx_logger.info(f"VDS shape set to {vds_shape}.")

        with h5py.File(self.filename, "r+") as nxs:
            if "jungfrau" in self.detector.detector_params.description.lower():
                jungfrau_vds_writer(
                    nxs,
                    (self.tot_num_imgs, *self.detector.detector_params.image_size),
                    data_type=vds_dtype,
                )
            else:
                image_vds_writer(
                    nxs,
                    (self.tot_num_imgs, *self.detector.detector_params.image_size),
                    start_index=vds_offset,
                    vds_shape=vds_shape,
                    data_type=vds_dtype,
                )
            if clean_up is True:
                nxmx_logger.warning("Starting clean up of unused links.")
                clean_unused_links(
                    nxs,
                    vds_shape=vds_shape,
                    start_index=vds_offset,
                )

            # If number of frames in the VDS is lower than the total, nimages in NXcollection should be overwritten to match this
            if vds_shape[0] < self.tot_num_imgs:
                del nxs["/entry/instrument/detector/detectorSpecific/nimages"]
                nxs["/entry/instrument/detector/detectorSpecific"].create_dataset(
                    "nimages", data=vds_shape[0]
                )


class EventNXmxFileWriter(NXmxFileWriter):
    """A class to generate NXmx-like NeXus files for event mode data."""

    def __init__(
        self,
        filename: Path | str,
        goniometer: Goniometer,
        detector: Detector,
        source: Source,
        beam: Beam,
        attenuator: Attenuator,
        axis_end_position: float | None = None,
    ):
        super().__init__(
            filename,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
            None,
        )
        self.end_pos = axis_end_position

    def write(
        self,
        start_time: datetime | str | None = None,
        write_mode: str = "x",
    ):
        """Write a NXmx-like NeXus file for event mode data collections.

        This method overrides the write() method of NXmxFileWriter, from which thsi class inherits.

        Args:
            start_time (datetime | str, optional): Collection estimated end time if available, in the format "%Y-%m-%dT%H:%M:%SZ".\
                Defaults to None.
            write_mode (str, optional): String indicating writing mode for the output NeXus file. Accepts any valid \
                h5py file opening mode. Defaults to "x".
        """
        # Get metafile
        # No data files, just link to meta
        metafile = super()._get_meta_file()

        module = self.detector.get_module_info()

        # Get scan info
        # Here no scan, just get (start, stop) from omega/phi as osc and None as transl
        osc, _ = self.goniometer.define_scan_axes_for_event_mode(self.end_pos)

        with h5py.File(self.filename, write_mode) as nxs:
            # NXentry and NXmx definition
            write_NXentry(nxs)

            # Write start time if known
            if start_time:
                write_NXdatetime(nxs, start_time, "start_time")

            # NXdata: entry/data
            write_NXdata(
                nxs,
                [metafile],
                self.goniometer.axes_list,
                "events",
                osc,
            )

            # NXinstrument: entry/instrument
            write_NXinstrument(
                nxs,
                self.beam,
                self.attenuator,
                self.source,
            )

            # NXdetector: entry/instrument/detector
            write_NXdetector(
                nxs,
                self.detector,
                meta=metafile,
            )

            # NXmodule: entry/instrument/detector/module
            write_NXdetector_module(
                nxs,
                module,
                self.detector.detector_params.image_size,
                self.detector.detector_params.pixel_size,
                beam_center=self.detector.beam_center,
            )

            # NXsource: entry/source
            write_NXsource(nxs, self.source)

            # NXsample: entry/sample
            write_NXsample(
                nxs,
                self.goniometer.axes_list,
                "events",
                osc,
                sample_depends_on=None,  # TODO
            )


class EDNXmxFileWriter(NXmxFileWriter):
    """A class to generate NXmx-like NeXus files for electron diffraction.

    Requires an additional argument:
        ED_coord_system (Dict): Definition of the current coordinate frame for ED. \
            It should at least contain the convention, origin and base vectors.
    """

    def __init__(
        self,
        filename: Path | str,
        goniometer: Goniometer,
        detector: Detector,
        source: Source,
        beam: Beam,
        attenuator: Attenuator,
        tot_num_imgs: int,
        ED_coord_system: Dict,
        convert_to_mcstas: bool = False,
    ):
        super().__init__(
            filename,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
            tot_num_imgs,
        )
        self.ED_coord_system = ED_coord_system
        self.convert_cs = convert_to_mcstas

    def _check_coordinate_frame(self):
        if self.convert_cs is True:
            nxmx_logger.warning(
                "All the vector/offset axis coordinates will be converted to mcstas."
            )
            mat = np.array(
                [
                    self.ED_coord_system["x"].vector,
                    self.ED_coord_system["y"].vector,
                    self.ED_coord_system["z"].vector,
                ]
            )
            # TODO: Need to redefine ED as {"x": Axis()} or something
            for ax in self.goniometer.axes_list:
                ax.vector = coord2mcstas(ax.vector, mat)
            for ax in self.detector.detector_axes:
                ax.vector = coord2mcstas(ax.vector, mat)
            self.detector.fast_axis = coord2mcstas(self.detector.fast_axis, mat)
            self.detector.slow_axis = coord2mcstas(self.detector.slow_axis, mat)

    def _get_data_filename(self) -> List[Path]:
        image_filename = f"{self.filename.stem}_data_000001.h5"
        return self.filename.parent / f"{image_filename}"

    def write(
        self,
        image_datafiles: List | None = None,
        data_entry_key: str = "/entry/data/data",
        start_time: datetime | str | None = None,
        write_mode: str = "x",
    ):
        """Write a NXmx-like NeXus file for electron diffraction.

        This method overrides the write() method of NXmxFileWriter, from which thsi class inherits.
        In particular, it performs a few checks on the coordinate frame of the input vectors \
        and then calls the writers for the relevant NeXus base classes.

        Args:
            image_datafiles (List | None, optional): List of image data files. If not passed, the program will look for \
                files with the stem_data_######.h5 in the target directory. Defaults to None.
            data_entry_key (str, optional): Dataset entry key in datafiles. Defaults to entry/data/data.
            start_time (datetime | str, optional): Collection estimated end time if available, in the format "%Y-%m-%dT%H:%M:%SZ".\
                Defaults to None.
            write_mode (str, optional): String indicating writing mode for the output NeXus file. Accepts any valid \
                h5py file opening mode. Defaults to "x".
        """
        # Get data files
        datafiles = image_datafiles if image_datafiles else [self._get_data_filename()]

        module = self.detector.get_module_info()

        # Scans
        osc, _ = self.goniometer.define_scan_from_goniometer_axes()

        # Deal with vecotrs/offsets/whatever if needed
        self._check_coordinate_frame()

        # Get the instrument name
        instrument_name = self.source.set_instrument_name()
        nxmx_logger.info(f"Instrument name will be set as {instrument_name}.")

        # NXcoordinate_system_set: /entry/coordinate_system_set
        base_vectors = {k: self.ED_coord_system.get(k) for k in ["x", "y", "z"]}

        with h5py.File(self.filename, write_mode) as nxs:
            # NXentry and NXmx definition
            write_NXentry(nxs)

            write_NXcoordinate_system_set(
                nxs,
                self.ED_coord_system["convention"],
                base_vectors,
                self.ED_coord_system["origin"],
            )

            if start_time:
                write_NXdatetime(nxs, start_time, "start_time")
                tot_exp_time = super()._get_collection_time()
                est_end = calculate_estimated_end_time(start_time, tot_exp_time)
                write_NXdatetime(nxs, est_end, "end_time_estimated")

            # NXdata: entry/data
            write_NXdata(
                nxs,
                datafiles,
                self.goniometer.axes_list,
                "images",
                osc,
                entry_key=data_entry_key,
            )

            # NXinstrument: entry/instrument
            write_NXinstrument(
                nxs,
                self.beam,
                self.attenuator,
                self.source,
                reset_instrument_name=True,
            )

            # NXdetector: entry/instrument/detector
            write_NXdetector(
                nxs,
                self.detector,
                self.tot_num_imgs,
            )

            # NXmodule: entry/instrument/detector/module
            write_NXdetector_module(
                nxs,
                module,
                self.detector.detector_params.image_size,
                self.detector.detector_params.pixel_size,
                beam_center=self.detector.beam_center,
            )

            # NXsource: entry/source
            write_NXsource(nxs, self.source)

            # NXsample: entry/sample
            write_NXsample(
                nxs,
                self.goniometer.axes_list,
                "images",
                osc,
            )

    def write_vds(
        self,
        vds_dtype: DTypeLike = np.uint16,
        writer_type: str = "dataset",
        data_entry_key: str = "/entry/data/data",
        datafiles: List[Path] | None = None,
    ):
        """Write a vds for electron diffraction.

        This method overrides the write_vds() method of NXmxFileWriter, from which thsi class inherits.
        In particular, if required it will write an external vds file instead of a dataset.

        Args:
            vds_dtype (DTypeLike, optional): The type of the input data. Defaults to np.uint16.
            writer_type (str, optional): Type of vds required. Defaults to "dataset".
            data_entry_key (str, optional): Dataset entry key in datafiles. Defaults to "/entry/data/data".
            datafiles ((List | None, optional): List of image data files. If not passed, the program will look for \
                files with the stem_######.h5 in the target directory. Defaults to None.
        """
        with h5py.File(self.filename, "r+") as nxs:
            if writer_type == "dataset":
                nxmx_logger.info(
                    "Writing vds dataset as /entry/data/data in nexus file."
                )
                image_vds_writer(
                    nxs,
                    (self.tot_num_imgs, *self.detector.detector_params.image_size),
                    data_type=vds_dtype,
                    entry_key=data_entry_key,
                )
            else:
                if not datafiles:
                    nxmx_logger.warning(
                        "No datafile list passed, vds file won't be written."
                    )
                    return
                nxmx_logger.info(
                    "Writing external vds file with link in /entry/data/data in nexus file."
                )
                vds_file_writer(
                    nxs,
                    datafiles,
                    (self.tot_num_imgs, *self.detector.detector_params.image_size),
                    vds_dtype,
                    data_entry_key,
                )



import json
import tempfile

import pytest

from nexgen.nxs_utils import Axis, TransformationType

axes_list = [
    Axis("omega", ".", TransformationType.ROTATION, (0, 0, -1)),
    Axis("sam_z", "omega", TransformationType.TRANSLATION, (0, 0, 1)),
    Axis("sam_y", "sam_z", TransformationType.TRANSLATION, (0, 1, 0)),
    Axis("sam_x", "sam_y", TransformationType.TRANSLATION, (1, 0, 0)),
]


@pytest.fixture
def dummy_chipmap_file():
    lines = [
        "01status    P3011       1\n",
        "02status    P3021       0\n",
        "03status    P3031       0\n",
        "04status    P3041       1\n",
    ]
    test_map_file = tempfile.NamedTemporaryFile(
        mode="w", suffix=".map", delete=False, encoding="utf-8"
    )
    with test_map_file as map:
        map.writelines(lines)
    yield test_map_file


@pytest.fixture
def dummy_chipmap_file_single_block():
    lines = [
        "01status    P3011       1\n",
        "02status    P3021       0\n",
        "03status    P3031       0\n",
        "04status    P3041       0\n",
    ]
    test_map_file = tempfile.NamedTemporaryFile(
        mode="w", suffix=".map", delete=False, encoding="utf-8"
    )
    with test_map_file as map:
        map.writelines(lines)
    yield test_map_file


@pytest.fixture
def dummy_chipmap_file_multi_block():
    lines = [
        "01status    P3011       1\n",
        "02status    P3021       1\n",
        "03status    P3031       1\n",
        "04status    P3041       1\n",
    ]
    test_map_file = tempfile.NamedTemporaryFile(
        mode="w", suffix=".map", delete=False, encoding="utf-8"
    )
    with test_map_file as map:
        map.writelines(lines)
    yield test_map_file


@pytest.fixture
def dummy_xml_file():
    test_xml = """<?xml version="1.0" encoding="UTF-8"?>
        <ExtendedCollectRequests>
            <fileinfo>
                <directory>/path/to/data</directory>
                <prefix>Filename</prefix>
            </fileinfo>
            <runNumber>1</runNumber>
            <oscillation_sequence>
                <start>-180.0</start>
                <range>2.0</range>
                <number_of_images>10</number_of_images>
                <exposure_time>0.2</exposure_time>
                <number_of_passes>1</number_of_passes>
            </oscillation_sequence>
            <sampleDetectorDistanceInMM>110.0</sampleDetectorDistanceInMM>
            <transmissionInPerCent>10.0</transmissionInPerCent>
            <kappa>0.0</kappa>
            <axisChoice>Omega</axisChoice>
            <twoTheta>10.0</twoTheta>
            <otherAxis>-90.0</otherAxis>
            <resolution>1</resolution>
        </ExtendedCollectRequests>
    """
    test_xml_file = tempfile.NamedTemporaryFile(suffix=".xml", delete=True)
    with open(test_xml_file.name, "w") as f:
        f.write(test_xml)
    yield test_xml_file


@pytest.fixture
def dummy_geometry_json():
    test_geometry = {
        "sample_omega": {
            "depends_on": ".",
            "ds_name": "omega",
            "vector": [-1, 0, 0],
            "location": "sample",
            "units": "deg",
            "type": "rotation",
        },
        "geometry": "mcstas",
        "sample_x": {
            "depends_on": "omega",
            "ds_name": "sam_x",
            "vector": [1, 0, 0],
            "location": "sample",
            "units": "mm",
            "type": "translation",
        },
        "phi": {
            "depends_on": "sam_x",
            "ds_name": "phi",
            "vector": [0, 0, -1],
            "location": "sample",
            "units": "deg",
            "type": "rotation",
        },
        "detector_z": {
            "depends_on": ".",
            "ds_name": "det_z",
            "vector": [0, 0, 1],
            "location": "detector",
            "units": "mm",
            "type": "translation",
        },
    }
    test_json_file = tempfile.NamedTemporaryFile(suffix=".json", delete=True)
    with open(test_json_file.name, "w") as f:
        json.dump(test_geometry, f, indent=4)
    yield test_json_file


@pytest.fixture
def dummy_detector_json():
    test_detector = {
        "eiger": {
            "pixel_size_units": "m",
            "description": "Eiger 2X",
            "sensor_material": "CdTe",
            "pixel_size": [7.5e-05, 7.5e-05],
            "size": [2162, 2068],
            "fast_dir": [0, 1, 0],
            "slow_dir": [-1, 0, 0],
        }
    }
    test_json_file = tempfile.NamedTemporaryFile(suffix=".json", delete=True)
    with open(test_json_file.name, "w") as f:
        json.dump(test_detector, f, indent=4)
    yield test_json_file



"""
Command line tool to generate NXmx-like NeXus files for Electron Diffraction.
"""
from __future__ import annotations

import argparse
import glob
import logging
import sys
from pathlib import Path

import freephil

from .. import log
from ..beamlines.ED_params import ED_coord_system
from ..beamlines.ED_singla_nxs import singla_nexus_writer
from ..nxs_utils import Axis, TransformationType
from . import config_parser, nexus_parser, version_parser

logger = logging.getLogger("nexgen.EDNeXusGeneratorCLI")

ED_phil = freephil.parse(
    """
    input {
      datafiles = None
        .type = path
        .help = "List of input data files."
      vds_writer = None *dataset file
        .type = choice
        .help = "If not None, write vds along with external link to data in NeXus file, or create _vds.h5 file."
      n_imgs = None
        .type = int
        .help = "Total number of images collected."
      convert_to_mcstas = False
        .type = bool
        .help = "Convert vectors to mcstas if required. Defaults to False."
    }

    include scope nexgen.command_line.nxs_phil.goniometer_scope

    include scope nexgen.command_line.nxs_phil.instrument_scope

    include scope nexgen.command_line.nxs_phil.detector_scope

    include scope nexgen.command_line.nxs_phil.module_scope

    include scope nexgen.command_line.nxs_phil.timestamp_scope

    include scope nexgen.command_line.nxs_phil.coord_system_scope
    """,
    process_includes=True,
)

usage = "%(prog)s <sub-command> [options]"
parser = argparse.ArgumentParser(
    usage=usage, description=__doc__, parents=[version_parser]
)
parser.add_argument("--debug", action="store_const", const=True)


def write_from_SINGLA(args):

    singla_nexus_writer(
        args.master_file,
        args.det_distance,
        args.exp_time,
        n_imgs=args.n_imgs if args.n_imgs else None,
        scan_axis=[args.axis_name, args.axis_start, args.axis_inc],
        beam_center=args.beam_center if args.beam_center else None,
        wavelength=args.wavelength if args.wavelength else None,
        outdir=args.output,
        start_time=args.start if args.start else None,
    )


def write_from_SINGLA_with_phil(args):
    cl = ED_phil.command_line_argument_interpreter()
    working_phil = ED_phil.fetch(cl.process_and_fetch(args.phil_args))
    params = working_phil.extract()

    if args.show_config:
        working_phil.show(attributes_level=args.attributes_level)
        sys.exit()

    datafiles = [
        Path(f).expanduser().resolve()
        for f in sorted(glob.glob(params.input.datafiles))
    ]

    logger.warning("Have you checked the coordinate system convention?\n")

    # If anything has been passed regarding the new coordinate system convention
    # overwrite the existing dictionary
    if params.coord_system.convention:
        logger.info(
            f"New coordinate system convention: {params.coord_system.convention}."
        )
        ED_coord_system["convention"] = params.coord_system.convention
    else:
        logger.info("The following convention will be applied:\n" f"{ED_coord_system}")

    if params.coord_system.origin:
        logger.info(
            f"New value for coordinate system found: {params.coord_system.origin}."
        )
        ED_coord_system["origin"] = tuple(params.coord_system.origin)

    if params.coord_system.vectors:
        from .cli_utils import split_arrays

        # Note: setting to coordinate frame to avoid any conversions.
        vectors = split_arrays(["x", "y", "z"], params.coord_system.vectors)
        logger.info(
            f"New vectors defined for {params.coord_system.convention} coordinate system."
        )
        ED_coord_system["x"] = Axis(
            "x", ".", TransformationType.TRANSLATION, vectors["x"]
        )
        ED_coord_system["y"] = Axis(
            "y", "x", TransformationType.TRANSLATION, vectors["y"]
        )
        ED_coord_system["z"] = Axis(
            "z", "y", TransformationType.TRANSLATION, vectors["z"]
        )

    # Source info from cli
    new_source = {
        "name": params.source.name,
        "facility_id": params.source.facility_id,
        "beamline": params.source.beamline_name,
        "probe": params.source.probe,
    }

    # Find scan axis in parsed values
    scan_idx = [
        n
        for n in range(len(params.goniometer.axes))
        if params.goniometer.types[n] == "rotation"
    ][0]
    scan_info = [
        params.goniometer.axes[scan_idx],
        params.goniometer.starts[scan_idx],
        params.goniometer.increments[scan_idx],
    ]

    singla_nexus_writer(
        args.master,
        params.detector.starts[0],
        params.detector.exposure_time,
        ED_coord_system,
        datafiles,
        params.input.convert_to_mcstas,
        n_imgs=params.input.n_imgs,
        scan_axis=scan_info,
        beam_center=params.detector.beam_center,
        wavelength=params.beam.wavelength,
        outdir=args.output,
        new_source_info=new_source,
        vds_writer=params.input.vds_writer,
    )


# Define some useful parsers
output_parser = argparse.ArgumentParser(add_help=False)
output_parser.add_argument(
    "-o",
    "--output",
    type=str,
    help="Output directory if different from location of data files.",
)

# Define subparsers
subparsers = parser.add_subparsers(
    help="Define what to do with input file depending on their origin.",
    required=True,
    dest="sub-command",
)
singla1_parser = subparsers.add_parser(
    "singla",
    description=("Trigger NeXus file writing for Singla data."),
    parents=[output_parser],
)
singla1_parser.add_argument(
    "master_file",
    type=str,
    help="HDF5 master file written by Singla detector.",
)
singla1_parser.add_argument(
    "det_distance",
    type=float,
    help="The sample-detector distance.",
)
singla1_parser.add_argument(
    "--axis-name",
    type=str,
    default="alpha",
    help="Rotation axis name",
)
singla1_parser.add_argument(
    "--axis-start",
    type=float,
    default=0.0,
    help="Rotation axis start position.",
)
singla1_parser.add_argument(
    "--axis-inc",
    type=float,
    default=0.0,
    help="Rotation axis increment.",
)
singla1_parser.add_argument(
    "-e",
    "--exp-time",
    type=float,
    required=True,
    help="Exposure time, in s.",
)
singla1_parser.add_argument(
    "-wl",
    "--wavelength",
    type=float,
    default=None,
    help="Incident beam wavelength, in A.",
)
singla1_parser.add_argument(
    "-bc",
    "--beam-center",
    type=float,
    nargs=2,
    help="Beam center (x,y) positions.",
)
singla1_parser.add_argument(
    "-n",
    "--n-imgs",
    type=int,
    help="Total number of images",
)
singla1_parser.add_argument(
    "--start", "--start-time", type=str, default=None, help="Collection start time."
)
singla1_parser.set_defaults(func=write_from_SINGLA)

singla2_parser = subparsers.add_parser(
    "singla-phil",
    description=("Trigger NeXus file writing for Singla data, using a phil parser."),
    parents=[nexus_parser, config_parser, output_parser],
)
singla2_parser.add_argument(
    "-m",
    "--master",
    type=str,
    required=True,
    help="HDF5 master file written by Singla detector.",
)
singla2_parser.set_defaults(func=write_from_SINGLA_with_phil)


def main():
    log.config()
    args = parser.parse_args()
    args.func(args)



"""
Object definition for Source, Beam and Attenuator
"""
from __future__ import annotations

from collections import namedtuple
from dataclasses import dataclass

from dataclasses_json import DataClassJsonMixin

# Describe facility
Facility = namedtuple("Facility", ("name", "short_name", "type", "id"))
Facility.__doc__ = """Facility description"""


class Source:
    """Source definition."""

    def __init__(
        self,
        beamline: str,
        facility: Facility = Facility(
            "Diamond Light Source",
            "DLS",
            "Synchrotron X-ray Source",
            None,
        ),
        probe: str | None = None,
    ):
        self.beamline = beamline

        self.name = facility.name
        self.short_name = facility.short_name
        self.facility_type = facility.type
        self.facility_id = facility.id

        self.probe = probe

    def __repr__(self) -> str:
        msg = f"Facility: {self.name} - {self.facility_type}. \n\t"
        msg += f"Beamline / instrument: {self.beamline} \n\t"
        if self.probe:
            msg += "Probe: {}"
        return f"Source information: \n\t{msg}"

    def set_instrument_name(self) -> str:
        """Set the instrument name from the details saved in source.

        If source type is not defined a priori, the function will assume it is a Synchrotron.
        If the facility_id is defined inside the source dictionary, that is the value that will be used.
        Naming tries to follow the recommended convention for NXmx:
        https://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_diffrn_source.type.html

        Returns:
            name (str): The name to write under '/entry/instrument/name'
        """
        facility_id = "DIAMOND" if self.facility_id is None else self.facility_id

        if (
            self.facility_type is None
            or "SYNCHROTRON" not in self.facility_type.upper()
        ):
            return f"{facility_id} {self.beamline}"
        else:
            return f"{facility_id} BEAMLINE {self.beamline}"

    def _generate_source_dict(self):
        source = {
            "name": self.name,
            "type": self.facility_type,
            "short_name": self.short_name,
            "beamline_name": self.beamline,
        }
        if self.probe:
            source["probe"] = self.probe
        if self.facility_id:
            source["facility_id"] = self.facility_id
        return source

    def to_dict(self):
        """Write source information to a dictionary."""
        return self._generate_source_dict()


@dataclass
class Beam(DataClassJsonMixin):
    """Beam definition."""

    wavelength: float
    flux: float | None = None


@dataclass
class Attenuator(DataClassJsonMixin):
    """Attenuator definition."""

    transmission: float



from unittest.mock import patch

import pytest

from nexgen.beamlines.SSX_Eiger_nxs import ssx_eiger_writer


@patch("nexgen.beamlines.SSX_Eiger_nxs.ssx_collect")
def test_writer_fails_for_Wrong_expt_type(fake_ssx_collect):
    with pytest.raises(ValueError):
        ssx_eiger_writer(
            "/path/to/file",
            "somefile",
            "i24",
            1,
            expt_type="aaa",
        )


@patch("nexgen.beamlines.SSX_Eiger_nxs.ssx_collect")
@patch("nexgen.beamlines.SSX_Eiger_nxs.log")
def test_writer_fails_for_beamline_not_i24_or_i19(fake_log, fake_ssx):
    with pytest.raises(ValueError):
        ssx_eiger_writer(
            "/path/to/file",
            "somefile",
            "i23",
            1,
        )



from nexgen.beamlines.GDAtools.ExtendedRequest import (
    ExtendedRequestIO,
    read_det_position_from_xml,
    read_scan_from_xml,
)
from nexgen.beamlines.GDAtools.GDAjson2params import JSONParamsIO


def test_get_coordinate_frame_from_json(dummy_geometry_json):
    assert JSONParamsIO(dummy_geometry_json.name).get_coordinate_frame() == "mcstas"


def test_get_gonio_axes_from_json(dummy_geometry_json):
    gonio_axes = JSONParamsIO(dummy_geometry_json.name).get_goniometer_axes_from_file()
    assert isinstance(gonio_axes, list) and len(gonio_axes) == 3
    assert gonio_axes[0].name == "omega" and gonio_axes[1].name == "sam_x"
    assert gonio_axes[2].name == "phi" and gonio_axes[2].depends == "sam_x"


def test_get_detector_axes_from_json(dummy_geometry_json):
    det_axes = JSONParamsIO(dummy_geometry_json.name).get_detector_axes_from_file()
    assert isinstance(det_axes, list) and len(det_axes) == 1
    assert det_axes[0].name == "det_z"


def test_get_fast_and_slow_dir_from_json(dummy_detector_json):
    fast, slow = JSONParamsIO(
        dummy_detector_json.name
    ).get_fast_and_slow_direction_vectors_from_file("eiger")
    assert (fast.x, fast.y, fast.z) == (0, 1, 0)
    assert (slow.x, slow.y, slow.z) == (-1, 0, 0)


def test_get_detector_params_from_json(dummy_detector_json):
    params = JSONParamsIO(dummy_detector_json.name).get_detector_params_from_file()
    assert params.description == "Eiger 2X"
    assert params.image_size == [2162, 2068]
    assert params.pixel_size == ["7.5e-05m", "7.5e-05m"]


def test_ExtendedRequestIO(dummy_xml_file):
    test_ecr = ExtendedRequestIO(dummy_xml_file.name)
    assert test_ecr.getAxisChoice() == "omega"
    assert test_ecr.getCollectionInfo() == ("/path/to/data", "Filename", "1")
    assert test_ecr.getTransmission() == 10.0
    assert test_ecr.getResolution() == 1.0


def test_read_scan_from_xml(dummy_xml_file):
    test_ecr = ExtendedRequestIO(dummy_xml_file.name)
    scan_axis, pos, num = read_scan_from_xml(test_ecr)
    assert scan_axis == test_ecr.getAxisChoice()
    assert num == 10
    assert isinstance(pos, dict) and len(pos) == 6  # gonio axes on I19-2
    assert pos["omega"] == (-180.0, -160.0, 2)
    assert pos["phi"] == (*2 * (test_ecr.getOtherAxis(),), 0.0)


def test_read_det_position_from_xml(dummy_xml_file):
    test_ecr = ExtendedRequestIO(dummy_xml_file.name)
    det_pos = read_det_position_from_xml(test_ecr, "Eiger")
    assert det_pos[0] == test_ecr.getTwoTheta()
    assert det_pos[1] == test_ecr.getSampleDetectorDistance()



"""
Experiment definitions for SSX collections:
    - extruder
    - fixed target
    - 3D grid scan
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np

from ..nxs_utils import Axis, TransformationType
from ..nxs_utils.ScanUtils import calculate_scan_points
from .beamline_utils import PumpProbe
from .SSX_chip import (
    Chip,
    compute_goniometer,
    fullchip_blocks_conversion,
    read_chip_map,
)

__all__ = ["run_extruder", "run_fixed_target", "run_3D_grid_scan"]

# Define a logger object
logger = logging.getLogger("nexgen.SSX.run_expt")


def run_extruder(
    goniometer_axes: List[Axis],
    num_imgs: int,
    pump_probe: PumpProbe,
    osc_axis: str = "omega",
) -> Tuple[List, Dict, Dict]:
    """Run the goniometer computations for an extruder experiment.

    Args:
        goniometer_axes (List[Axis]): List of goniometer axes for current beamline.
        num_imgs (int): Total number of images.
        pump_probe (PumpProbe): Pump probe parameters.
        osc_axis: Defines which axis is considered the "moving" one. Defaults to omega.

    Returns:
        Tuple[List, Dict, Dict]:
            goniometer_axes: updated goniometer_axes list with actual values from the scan.
            SCAN: dictionary with oscillation scan axis values.
            pump_info: updated pump probe information.
    """
    logger.debug("Running an extruder experiment.")

    logger.debug("All axes are fixed, setting increments to 0.0 and starts == ends.")
    for ax in goniometer_axes:
        # Sanity check that no increment is greater than 0.0
        if ax.transformation_type == "rotation":
            ax.increment = 0.0

    # Identify the "oscillation axis"
    osc_idx = [n for n, ax in enumerate(goniometer_axes) if ax.name == osc_axis][0]
    goniometer_axes[osc_idx].num_steps = num_imgs

    # Calculate scan
    logger.debug(
        "Getting 'oscillation scan': roation axis not moving, same value for each image as."
    )
    SCAN = calculate_scan_points(
        goniometer_axes[osc_idx], rotation=True, tot_num_imgs=num_imgs
    )

    pump_info = pump_probe.to_dict()
    logger.debug("Removing pump_repeat from pump probe necessary information.")
    pump_info.pop("pump_repeat")

    return goniometer_axes, SCAN, pump_info


def run_fixed_target(
    goniometer_axes: List[Axis],
    chip_info: Dict[str, List],
    chipmap: Path | str,
    pump_probe: PumpProbe,
    scan_axes: List[str, str] = ["sam_y", "sam_x"],
) -> Tuple[Dict, Dict]:
    """Run the goniometer computations for a fixed-target experiment.

    Args:
        goniometer_axes (List[Axis]): List of goniometer axes for current beamline.
        chip_info (Dict[str, List]): General information about the chip: number and size of blocks, \
            size and step of each window, start positions, number of exposures.
        chipmap (Path | str): Path to .map file. If None is passed, assumes a fullchip.
        pump_probe (PumpProbe): Pump probe parameters.
        scan_axes (List[str, str], optional): List of scan axes, in order slow,fast. \
            Defaults to ["sam_y", "sam_x"].

    Raises:
        ValueError: If one or both of the axes names passed as input are not part of the goniometer axes.
        ValueError:if chip_info hasn't been passed or is an empty dictionary.

    Returns:
        Tuple[Dict, Dict]:
            SCAN: Dictionary with grid scan values.
            pump_info: Updated pump probe information.
    """
    logger.info("Running a fixed target experiment.")

    # Check that the axes for the scan make sense
    check_list = [n for n, ax in enumerate(goniometer_axes) if ax.name in scan_axes]
    if len(check_list) < len(scan_axes):
        raise ValueError(
            "Axis not found in the list of goniometer axes. Please check your input."
            f"Goniometer axes: {goniometer_axes}. Looking for {scan_axes}."
        )

    # Check that the chip dict has been passed, raise error if not
    if not chip_info:
        logger.error("No chip_dict found.")
        raise ValueError(
            "No information about the FT chip has been passed. \
            Impossible to determine scan parameters. NeXus file won't be written."
        )

    # Define chip
    chip = Chip(
        "fastchip",
        num_steps=[chip_info["X_NUM_STEPS"][1], chip_info["Y_NUM_STEPS"][1]],
        step_size=[chip_info["X_STEP_SIZE"][1], chip_info["Y_STEP_SIZE"][1]],
        num_blocks=[chip_info["X_NUM_BLOCKS"][1], chip_info["Y_NUM_BLOCKS"][1]],
        block_size=[chip_info["X_BLOCK_SIZE"][1], chip_info["Y_BLOCK_SIZE"][1]],
        start_pos=[
            chip_info["X_START"][1],
            chip_info["Y_START"][1],
            chip_info["Z_START"][1],
        ],
    )

    # Read chip map
    blocks = read_chip_map(
        chipmap,
        chip.num_blocks[0],
        chip.num_blocks[1],
    )

    # Workaround for eg. I19 Eiger which saves an increment for phi/omega in meta file.
    for ax in goniometer_axes:
        if ax.transformation_type == "rotation":
            ax.increment = 0.0

    # Calculate scan start positions on chip
    if list(blocks.values())[0] == "fullchip":
        logger.info("Full chip: all the blocks will be scanned.")
        starts = compute_goniometer(chip, full=True, ax1=scan_axes[0], ax2=scan_axes[1])
        starts = fullchip_blocks_conversion(starts, chip)
    else:
        logger.info(f"Scanning blocks: {list(blocks.keys())}.")
        starts = compute_goniometer(chip, blocks, ax1=scan_axes[0], ax2=scan_axes[1])

    # Create two temporary axes to be used for scan calculations
    axis1 = Axis(scan_axes[0], "", TransformationType.TRANSLATION, (0, 0, 0))
    axis2 = Axis(scan_axes[1], "", TransformationType.TRANSLATION, (0, 0, 0))

    # Iterate over blocks to calculate scan points
    SCAN = {axis1.name: np.array([]), axis2.name: np.array([])}
    for k, v in starts.items():
        axis1.start_pos = v[axis1.name]
        axis1.increment = chip.step_size[1] * v["direction"]
        axis1.num_steps = chip.num_steps[1]
        axis2.start_pos = v[axis2.name]
        axis2.increment = chip.step_size[0]
        axis2.num_steps = chip.num_steps[0]

        logger.debug(
            f"Current block: {k}\n"
            f"{axis1.name} start: {v[axis1.name]} \n"
            f"{axis2.name} start: {v[axis2.name]} \n"
            f"Scan direction: {v['direction']} \n"
        )

        _scan = calculate_scan_points(axis1, axis2)
        SCAN[axis1.name] = np.append(SCAN[axis1.name], np.round(_scan[axis1.name], 3))
        SCAN[axis2.name] = np.append(SCAN[axis2.name], np.round(_scan[axis2.name], 3))

    # Check the number of exposures per window
    N = int(chip_info["N_EXPOSURES"][1])
    if N > 1:
        # Repeat each position N times
        SCAN = {k: [val for val in v for _ in range(N)] for k, v in SCAN.items()}
    logger.info(f"Each position has been collected {N} times.")
    logger.info(f"Pump repeat setting: {chip_info['PUMP_REPEAT'][1]}.")
    pump_info = pump_probe.to_dict()
    pump_info["n_exposures"] = N

    return SCAN, pump_info


def run_3D_grid_scan(
    goniometer_axes: List[Axis],
    chip_info: Dict[str, List],
    chipmap: Path | str,
    pump_probe: PumpProbe,
    osc_axis: str = "omega",
) -> Tuple[Dict]:
    """_summary_

    Args:
        goniometer_axes (List[Axis]): _description_
        chip_info (Dict[str, List]): _description_
        chipmap (Path | str): _description_
        pump_probe (PumpProbe): _description_
        osc_axis (str, optional): _description_

    Returns:
        Tuple[Dict]:
            OSC: dictionary with oscillation scan axis values
            TRANSL: dictionary with grid scan values
            pump_info: updated pump probe information
    """
    logger.info("Running a 3D grid scan experiment.")

    N = int(chip_info["N_EXPOSURES"][1])

    pump_info = pump_probe.to_dict()
    pump_info["repeat"] = int(chip_info["PUMP_REPEAT"][1])
    pump_info["n_exposures"] = N
    return None, None, pump_info



"""
Create a NeXus file for time-resolved collections on I19-2.
"""
from __future__ import annotations

import logging
from collections import namedtuple
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Sequence, Tuple, Union

import h5py
import numpy as np
from pydantic import BaseModel

from .. import log
from ..nxs_utils import (
    Attenuator,
    Beam,
    Detector,
    EigerDetector,
    Goniometer,
    Source,
    TristanDetector,
)
from ..nxs_utils.ScanUtils import calculate_scan_points, identify_osc_axis
from ..nxs_write.NXmxWriter import EventNXmxFileWriter, NXmxFileWriter
from ..tools.Metafile import DectrisMetafile
from ..tools.MetaReader import define_vds_data_type, update_axes_from_meta
from ..utils import find_in_dict, get_iso_timestamp, get_nexus_filename
from .beamline_utils import collection_summary_log


class CollectionParams(BaseModel):
    """Parameters passed as input from the beamline.

    Args:
        metafile: Path to _meta.h5 file.
        detector_name: Name of the detector in use for current experiment.
        exposure_time: Exposure time, in s.
        beam_center: Beam center (x,y) position, in pixels.
        wavelength: Incident beam wavelength, in A.
        transmission: Attenuator transmission, in %.
        tot_num_images: Total number of frames in a collection.
        scan_axis: Rotation scan axis. Must be passed for Tristan.
    """

    metafile: Union[Path, str]
    detector_name: str
    exposure_time: float
    beam_center: Sequence[float]
    wavelength: Optional[float]
    transmission: Optional[float]
    tot_num_images: Optional[int]
    scan_axis: Optional[str]


class ExperimentTypeError(Exception):
    pass


# Define a logger object
logger = logging.getLogger("nexgen.I19-2_NeXus")


# Useful axis definitions
axes = namedtuple("axes", ("id", "start", "inc", "end"), defaults=(None, 0.0, 0.0, 0.0))
axes.__doc__ = """Goniometer axis name, start and end position, increment."""
det_axes = namedtuple("det_axes", ("id", "start"), defaults=(None, 0.0))
det_axes.__doc__ = """Detector axis name and position."""


def tristan_writer(
    master_file: Path,
    TR: CollectionParams,
    timestamps: Tuple[str, str] = (None, None),
    axes_pos: List[axes] = None,
    det_pos: List[det_axes] = None,
):
    """
    A function to call the nexus writer for Tristan 10M detector.

    Args:
        master_file (Path): Path to nexus file to be written.
        TR (CollectionParams): Parameters passed from the beamline.
        timestamps (Tuple[str, str], optional): Collection start and end time. Defaults to (None, None).
        axes_pos (List[axes], optional): List of (axis_name, start, end) values for the \
            goniometer, passed from command line. Defaults to None.
        det_pos (List[det_axes], optional): List of (axis_name, start) values for the \
            detector, passed from command line. Defaults to None.
    """
    source = Source("I19-2")
    from .I19_2_params import I19_2Tristan as axes_params

    # Define Tristan params
    tristan_params = TristanDetector("Tristan 10M", (3043, 4183))

    # Define Goniometer axes
    gonio_axes = axes_params.gonio
    # Define Detector
    det_axes = axes_params.det_axes

    # Update axes
    # Goniometer
    end_pos = None
    for gax in axes_pos:
        idx = [n for n, ax in enumerate(gonio_axes) if ax.name == gax.id][0]
        gonio_axes[idx].start_pos = gax.start
        if gax.start != gax.end:
            end_pos = gax.end

    # Detector
    for dax in det_pos:
        idx = [n for n, ax in enumerate(det_axes) if ax.name == dax.id][0]
        det_axes[idx].start_pos = dax.start

    # Identify scan axis and calculate scan range
    scan_axis = TR.scan_axis if TR.scan_axis else "phi"
    scan_idx = [n for n, ax in enumerate(gonio_axes) if ax.name == scan_axis][0]
    if not end_pos:
        end_pos = gonio_axes[scan_idx].end_pos
    OSC = {scan_axis: (gonio_axes[scan_idx].start_pos, end_pos)}

    # Define Detector
    detector = Detector(
        tristan_params,
        det_axes,
        TR.beam_center,
        TR.exposure_time,
        [axes_params.fast_axis, axes_params.slow_axis],
    )

    # Define Goniometer
    goniometer = Goniometer(gonio_axes, OSC)

    # Define beam and attenuator
    attenuator = Attenuator(TR.transmission)
    beam = Beam(TR.wavelength)

    collection_summary_log(
        logger,
        gonio_axes,
        [scan_axis],
        detector,
        attenuator,
        beam,
        source,
        timestamps,
    )

    # Write
    try:
        EventFileWriter = EventNXmxFileWriter(
            master_file,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
        )
        EventFileWriter.write(start_time=timestamps[0])
        if timestamps[1]:
            EventFileWriter.update_timestamps(timestamps[1], "end_time")
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        raise


def eiger_writer(
    master_file: Path,
    TR: CollectionParams,
    timestamps: Tuple[str, str] = (None, None),
    use_meta: bool = False,
    n_frames: int | None = None,
    axes_pos: List[axes] = None,
    det_pos: List[det_axes] = None,
    vds_offset: int = 0,
):
    """
    A function to call the NXmx nexus file writer for Eiger 2X 4M detector.
    If use_meta is set to False, axes_pos and det_pos become required arguments. Otherwise, \
    axes_pos and det_pos can be None but the code requires the information contained inside \
    the meta file to work correctly.

    Args:
        master_file (Path): Path to nexus file to be written.
        TR (CollectionParams): Parameters passed from the beamline.
        timestamps (Tuple[str, str], optional): Collection start and end time. Defaults to (None, None).
        use_meta (bool, optional): If True, metadata such as axes positions, wavelength etc. \
            will be updated using the meta.h5 file. Defaults to False.
        num_frames (int, optional): Number of images for the nexus file. Not necessary if same as the \
            tot_num_images from the CollectionParameters. If different, the VDS will onlu contain the \
            number of frames specified here. Defaults to None.
        axes_pos (List[axes], optional): List of (axis_name, start, inc) values for the \
            goniometer, passed from command line. Defaults to None.
        det_pos (List[det_axes], optional): List of (axis_name, start) values for the \
            detector, passed from command line. Defaults to None.
        vds_offset (int, optional): Start index for the vds writer. Defaults to 0.

    Raises:
        ValueError: If use_meta is set to False but axes_pos and det_pos haven't been passed.
        IOError: If the axes positions can't be read from the metafile (missing config or broken links).
    """
    if not use_meta:
        if axes_pos is None or det_pos is None:
            logger.error(
                """
                If not using the meta file, please pass the complete axis information for goniometer
                and/or detector.
                """
            )
            raise ValueError("Missing at least one of axes_pos or det_pos.")
        if n_frames is None and TR.tot_num_images is None:
            logger.error(
                """
                If not using the meta file, please make sure either the total number of images is passed \
                or the number of frames has been passed. These values could be the same for a standard \
                collection, or different if the vds needs to only point to part of the dataset.
                """
            )
            raise ValueError("Missing number of images.")

    source = Source("I19-2")
    from .I19_2_params import I19_2Eiger as axes_params

    # Define Eiger 4M params
    eiger_params = EigerDetector(
        "Eiger 2X 4M",
        (2162, 2068),
        "CdTe",
        50649,
        -1,
    )

    # Read some parameters
    transmission = TR.transmission if TR.transmission else None
    wl = TR.wavelength
    beam_center = TR.beam_center

    # Define Goniometer axes
    gonio_axes = axes_params.gonio
    # Define Detector
    det_axes = axes_params.det_axes

    # Update axes
    if use_meta:
        logger.info("User requested to update metadata using meta file.")
        with h5py.File(TR.metafile, "r", libver="latest", swmr=True) as mh:
            meta = DectrisMetafile(mh)
            TR.tot_num_images = meta.get_full_number_of_images()
            logger.info(
                f"Total number of images for this collection found in meta file: {TR.tot_num_images}."
            )
            if not n_frames:
                n_frames = TR.tot_num_images
                logger.info(
                    "No specific numnber of frames requested, VDS will contain the full dataset."
                )
            vds_dtype = define_vds_data_type(meta)
            update_axes_from_meta(
                meta, gonio_axes, osc_axis=TR.scan_axis, use_config=True
            )
            update_axes_from_meta(meta, det_axes)
            # WARNING.det_z not in _dectris, but det_distance is. Getting that.

            logger.info(
                "Goniometer and detector axes positions have been updated with values from the meta file."
            )
            if TR.wavelength is None:
                logger.info(
                    "Wavelength hasn't been passed by user. Looking for it in the meta file."
                )
                wl = meta.get_wavelength()
            if TR.beam_center is None:
                logger.info(
                    "Beam center position has't been passed by user. Looking for it in the meta file."
                )
                beam_center = meta.get_beam_center()
    else:
        logger.info(
            "Not using meta file to update metadata, only the external links will be set up."
        )
        vds_dtype = np.uint32
        # Update axes
        # Goniometer
        for gax in axes_pos:
            idx = [n for n, ax in enumerate(gonio_axes) if ax.name == gax.id][0]
            gonio_axes[idx].start_pos = gax.start
            if gax.inc != 0.0:
                gonio_axes[idx].increment = gax.inc

        # Detector
        for dax in det_pos:
            idx = [n for n, ax in enumerate(det_axes) if ax.name == dax.id][0]
            det_axes[idx].start_pos = dax.start
            logger.info(
                "Goniometer and detector axes positions have been updated with values passed by the user."
            )

        if not n_frames:
            n_frames = TR.tot_num_images
        if not TR.tot_num_images:
            TR.tot_num_images = n_frames
            logger.warning(
                """
                As the total number of images was not set in the collection parameters, it has been set to \
                the requested number of frames.
                """
            )

    scan_axis = identify_osc_axis(gonio_axes)
    # Check that found scan_axis matches
    if scan_axis != TR.scan_axis:
        logger.warning(
            f"Scan axis {scan_axis} found different from requested one {TR.scan_axis}."
            f"Defaulting to {TR.scan_axis}. If wrong please check meta file."
        )
        scan_axis = TR.scan_axis
    scan_idx = [n for n, ax in enumerate(gonio_axes) if ax.name == scan_axis][0]
    gonio_axes[scan_idx].num_steps = n_frames
    OSC = calculate_scan_points(
        gonio_axes[scan_idx],
        rotation=True,
        tot_num_imgs=n_frames,
    )

    # Define beam and attenuator
    attenuator = Attenuator(transmission)
    beam = Beam(wl)

    # Define Detector
    detector = Detector(
        eiger_params,
        det_axes,
        beam_center,
        TR.exposure_time,
        [axes_params.fast_axis, axes_params.slow_axis],
    )

    # Define Goniometer
    goniometer = Goniometer(gonio_axes, OSC)

    collection_summary_log(
        logger,
        goniometer,
        detector,
        attenuator,
        beam,
        source,
        timestamps,
    )

    # Write
    try:
        image_filename = TR.metafile.as_posix().replace("_meta.h5", "")
        NXmx_writer = NXmxFileWriter(
            master_file,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
            TR.tot_num_images,
        )
        NXmx_writer.write(image_filename=image_filename, start_time=timestamps[0])
        NXmx_writer.write_vds(
            vds_offset=vds_offset,
            vds_shape=(n_frames, *detector.detector_params.image_size),
            vds_dtype=vds_dtype,
        )
        if timestamps[1]:
            NXmx_writer.update_timestamps(timestamps[1], "end_time")
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        raise


def nexus_writer(
    meta_file: Path | str,
    detector_name: str,
    exposure_time: float,
    scan_axis: str = "phi",
    start_time: datetime | None = None,
    stop_time: datetime | None = None,
    **params,
):
    """
    Gather all parameters from the beamline and call the NeXus writers.

    Args:
        meta_file (Path | str): Path to _meta.h5 file.
        detector_name (str): Detector in use.
        exposure_time (float): Exposure time, in s.
        scan_axis (str, optional): Name of the oscillation axis. Defaults to phi.
        start_time (datetime, optional): Experiment start time. Defaults to None.
        stop_time (datetime, optional): Experiment end time. Defaults to None.

    Keyword Args:
        n_imgs (int): Total number of images to be collected.
        transmission (float): Attenuator transmission, in %.
        wavelength (float): Wavelength of incident beam, in A.
        beam_center (List[float, float]): Beam center position, in pixels.
        gonio_pos (List[axes]): Name, start and end positions \
            of the goniometer axes.
        det_pos (List[det_axes]): Name, start and end positions \
            of detector axes.
        outdir (str): Directory where to save the file. Only specify if different \
            from meta_file directory.
        serial (bool): Specify whether it's a serial crystallography dataset.
        det_dist (float): Distance between sample and detector, in mm.
        use_meta (bool): For Eiger, if True use metadata from meta.h5 file. Otherwise \
            will require all other information to be passed manually.
    """
    if find_in_dict("serial", params) and params["serial"] is True:
        raise ExperimentTypeError(
            "This is writer is not enabled for ssx collections."
            "Pleas look into SSX_Eiger or SSX_Tristan for this functionality."
        )

    TR = CollectionParams(
        metafile=Path(meta_file).expanduser().resolve(),
        detector_name=detector_name.lower(),
        exposure_time=exposure_time,
        beam_center=params["beam_center"]
        if find_in_dict("beam_center", params)
        else (0, 0),
        wavelength=params["wavelength"] if find_in_dict("wavelength", params) else None,
        transmission=params["transmission"]
        if find_in_dict("transmission", params)
        else None,
        tot_num_images=params["n_imgs"] if find_in_dict("n_imgs", params) else None,
        scan_axis=scan_axis,
    )

    # Check that the new NeXus file is to be written in the same directory
    if find_in_dict("outdir", params) and params["outdir"]:
        wdir = Path(params["outdir"]).expanduser().resolve()
    else:
        wdir = TR.metafile.parent

    # Define a file handler
    logfile = wdir / "I19_2_nxs_writer.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info("NeXus file writer for beamline I19-2 at DLS.")
    logger.info(f"Detector in use for this experiment: {TR.detector_name}.")
    logger.info(f"Current collection directory: {TR.metafile.parent}")

    # Add some information to logger
    logger.info("Creating a NeXus file for %s ..." % TR.metafile.name)
    # Get NeXus filename
    master_file = get_nexus_filename(TR.metafile)
    master_file = wdir / master_file.name
    logger.info("NeXus file will be saved as %s" % master_file)

    # Get timestamps in the correct format if they aren't already
    start_time = start_time.strftime("%Y-%m-%dT%H:%M:%S") if start_time else None
    stop_time = stop_time.strftime("%Y-%m-%dT%H:%M:%S") if stop_time else None
    timestamps = (
        get_iso_timestamp(start_time),
        get_iso_timestamp(stop_time),
    )

    if not find_in_dict("gonio_pos", params):
        params["gonio_pos"] = None
    if not find_in_dict("det_pos", params):
        params["det_pos"] = None

    if "tristan" in TR.detector_name.lower():
        if params["gonio_pos"] is None or params["det_pos"] is None:
            logger.error("Please pass the axes positions for a Tristan collection.")
            raise ValueError(
                "Missing goniometer and/or detector axes information for tristan collection"
            )
        if TR.scan_axis is None:
            logger.warning(
                "No scan axis has been specified. Phi will be set as default."
            )

    if not find_in_dict("use_meta", params):
        # If by any chance not passed, assume False
        params["use_meta"] = False

    if params["use_meta"] is True:
        params["gonio_pos"] = None
        params["det_pos"] = None
    else:
        if not find_in_dict("n_imgs", params) and "eiger" in TR.detector_name:
            raise ValueError(
                """
                Missing input parameter n_imgs. \n
                For an Eiger collection, if meta file is to be ignored, the number of images to
                be collected has to be passed to the writer.
                """
            )
        if TR.beam_center == (0, 0):
            logger.warning(
                """
                Beam centre was not passed to the writer.
                As it won't be updated from the meta file, it will be set to (0, 0).
                """
            )

    if "eiger" in TR.detector_name:
        if not find_in_dict("n_imgs", params):
            params["n_imgs"] = None
        eiger_writer(
            master_file,
            TR,
            timestamps,
            params["use_meta"],
            params["n_imgs"],
            params["gonio_pos"],
            params["det_pos"],
        )
    elif "tristan" in TR.detector_name:
        tristan_writer(
            master_file, TR, timestamps, params["gonio_pos"], params["det_pos"]
        )



"""
Tools for copying the metadata from Tristan NeXus files.
"""
from __future__ import annotations

import logging
from pathlib import Path

import h5py
import numpy as np

from ..nxs_write.NXclassWriters import write_NXnote
from ..nxs_write.write_utils import create_attributes
from .copy_utils import (
    check_and_fix_det_axis,
    compute_ssx_axes,
    convert_scan_axis,
    get_nexus_tree,
    identify_tristan_scan_axis,
)

tristan_logger = logging.getLogger("nexgen.CopyTristanNeXus")


def single_image_nexus(
    data_file: Path | str,
    tristan_nexus: Path | str,
    write_mode: str = "x",
    pump_probe_bins: int = None,
) -> str:
    """
    Create a NeXus file for a single-image or a stationary pump-probe dataset.

    Copy the nexus tree from the original NeXus file for a collection on Tristan
    detector. In the case of a single image, the input scan_axis is a (start, stop) tuple where start and
    stop have the same value, for a pump-probe experiment the values might differ for some older datasets.
    The scan_axis in the new file will therefore be one single number, equal to the "start".

    Args:
        data_file (Path | str): String or Path pointing to the HDF5 file containing the newly binned images.
        tristan_nexus (Path | str): String or Path pointing to the input NeXus file with experiment metadata to be copied.
        write_mode (str, optional): String indicating writing mode for the output NeXus file.  Accepts any valid
                        h5py file opening mode. Defaults to "x".
        pump_probe_bins (int, optional): If the NeXus file is be linked to a static pump-probe image stack, pass the number
                        of images the events have been binned into. Deafults to None.

    Returns:
        nxs_filename (str): The name of the output NeXus file.
    """
    data_file = Path(data_file).expanduser().resolve()
    tristan_nexus = Path(tristan_nexus).expanduser().resolve()
    nxs_filename = data_file.parent / f"{data_file.stem}.nxs"
    with h5py.File(tristan_nexus, "r") as nxs_in, h5py.File(
        nxs_filename, write_mode
    ) as nxs_out:
        # Copy the whole tree except for nxdata
        nxentry = get_nexus_tree(nxs_in, nxs_out)
        # Create nxdata group
        nxdata = nxentry.create_group("data")
        # Add link to data
        nxdata["data"] = h5py.ExternalLink(data_file.name, "data")
        # Compute and write axis information
        ax, ax_attr = identify_tristan_scan_axis(nxs_in)
        if ax:
            create_attributes(
                nxdata,
                ("NX_class", "axes", "signal", ax + "_indices"),
                (
                    "NXdata",
                    ax,
                    "data",
                    [
                        0,
                    ],
                ),
            )
            try:
                ax_range = nxs_in["entry/data"][ax][0]
            except ValueError:
                # Some early Tristan data from before March 2021, where the goniometer
                # was not moved during the data collection, record the rotation axis
                # position as a scalar.
                ax_range = nxs_in["entry/data"][ax][()]
            if pump_probe_bins is not None:
                ax_range = np.repeat(ax_range, pump_probe_bins)
                nxdata.create_dataset(ax, data=ax_range)
            else:
                nxdata.create_dataset(ax, data=np.array([ax_range]))
            # Write the attributes
            for key, value in ax_attr.items():
                nxdata[ax].attrs.create(key, value)
            # Now fix all other instances of scan_axis in the tree
            nxsample = nxentry["sample"]
            convert_scan_axis(nxsample, nxdata, ax)

    return nxs_filename.as_posix()


def multiple_images_nexus(
    data_file: Path | str,
    tristan_nexus: Path | str,
    write_mode: str = "x",
    osc: float = None,
    nbins: int = None,
) -> str:
    """
    Create a NeXus file for a multiple-image dataset or multiple image sequences from a pump-probe collection.

    Copy the nexus tree from the original NeXus file for a collection on Tristan detector. There are two main applications for this function.
    In the first case, multiple images from a rotation collection have been binned and the scan_axis to be found in the input file is a (start, stop) tuple.
    The scan_axis in the new file will therefore be a list of angles. Osc and num_bins are mutually exclusive arguments to work out the scan_axis list.
    In the second case, multiple images from a 2D grid scan collection have been binned. The "rotation" scan_axis is still to be found in the input file as a
    (start, stop) tuple - although in this instance the two values should coincide. The values for the "translation" scan axes instead can be worked out from
    the chipmap dictionary, saved as a Unicode string inside the original NeXus file during collection, and nbins. It should be noted that passing osc in this
    case will raise an error and exit.

    Args:
        data_file (Path | str): String or Path pointing to the HDF5 file containing the newly binned images.
        tristan_nexus (Path | str): String or Path pointing to the input NeXus file with experiment metadata to be copied.
        write_mode (str, optional): String indicating writing mode for the output NeXus file.  Accepts any valid
                        h5py file opening mode. Defaults to "x".
        osc (float, optional): Oscillation angle (degrees). Defaults to None.
        nbins (int, optional): Number of binned images. Defaults to None.

    Raises:
        ValueError: When osc has been passed instead of nbins for a grid scan collection.
        ValueError: When both osc and nbins have been passed for a rotation collection. The two values are mutually exclusive.
        ValueError: When neither osc nor nbins has been passed. It won't be possible to calculate the scan range without at least one of them.

    Returns:
        nxs_filename (str): The name of the output NeXus file.
    """
    data_file = Path(data_file).expanduser().resolve()
    tristan_nexus = Path(tristan_nexus).expanduser().resolve()
    nxs_filename = data_file.parent / f"{data_file.stem}.nxs"
    with h5py.File(tristan_nexus, "r") as nxs_in, h5py.File(
        nxs_filename, write_mode
    ) as nxs_out:
        # Copy the whole tree except for nxdata
        nxentry = get_nexus_tree(nxs_in, nxs_out)
        # Create nxdata group
        nxdata = nxentry.create_group("data")
        # Add link to data
        nxdata["data"] = h5py.ExternalLink(data_file.name, "data")
        # Compute and write axis information
        ax, ax_attr = identify_tristan_scan_axis(nxs_in)
        if ax:
            create_attributes(
                nxdata,
                ("NX_class", "axes", "signal", ax + "_indices"),
                (
                    "NXdata",
                    ax,
                    "data",
                    [
                        0,
                    ],
                ),
            )
            try:
                (start, stop) = nxs_in["entry/data"][ax][()]
            except (TypeError, ValueError):
                # Some early Tristan data from before March 2021, where the goniometer
                # was not moved during the data collection, record the rotation axis
                # position as a scalar.
                start = stop = nxs_in["entry/data"][ax][()]

            if osc and nbins:
                raise ValueError(
                    "osc and nbins are mutually exclusive, "
                    "please pass only one of them."
                )
            elif osc:
                ax_range = np.arange(start, stop, osc)
            elif nbins:
                ax_range = np.linspace(start, stop, nbins + 1)[:-1]
            else:
                raise ValueError(
                    "Impossible to calculate scan_axis, "
                    "please pass either osc or nbins."
                )

            nxdata.create_dataset(ax, data=ax_range)
            # Write the attributes
            for key, value in ax_attr.items():
                nxdata[ax].attrs.create(key, value)
            # Now fix all other instances of scan_axis in the tree
            nxsample = nxentry["sample"]
            convert_scan_axis(nxsample, nxdata, ax, ax_range)

    return nxs_filename.as_posix()


def serial_images_nexus(
    data_file: Path | str,
    tristan_nexus: Path | str,
    nbins: int,
    write_mode: str = "x",
) -> str:
    """
    Create a NeXus file for a serial collection.

    Args:
        data_file (Path | str): String or Path pointing to the HDF5 file containing the newly binned images.
        tristan_nexus (Path | str): String or Path pointing to the input NeXus file with experiment metadata to be copied.
        nbins (int): Number of binned images.
        write_mode (str, optional): String indicating writing mode for the output NeXus file.  Accepts any valid
                        h5py file opening mode. Defaults to "x".

    Returns:
        str: _description_
    """
    data_file = Path(data_file).expanduser().resolve()
    tristan_nexus = Path(tristan_nexus).expanduser().resolve()
    nxs_filename = data_file.parent / f"{data_file.stem}.nxs"
    with h5py.File(tristan_nexus, "r") as nxs_in, h5py.File(
        nxs_filename, write_mode
    ) as nxs_out:
        # Copy the whole tree except for nxdata and nxnote (which is where chip info is... or the pump for the old ones)
        nxentry = get_nexus_tree(nxs_in, nxs_out, skip_obj=["NXdata", "NXnote"])
        # Create nxdata group
        nxdata = nxentry.create_group("data")
        # Add link to data
        nxdata["data"] = h5py.ExternalLink(data_file.name, "data")
        # Compute and write axis information
        ax, ax_attr = identify_tristan_scan_axis(nxs_in)
        if ax:
            create_attributes(
                nxdata,
                ("NX_class", "axes", "signal", ax + "_indices"),
                (
                    "NXdata",
                    ax,
                    "data",
                    [
                        0,
                    ],
                ),
            )

            (start, stop) = nxs_in["entry/data"][ax][()]

            # Compute the scan points
            rot_ax, transl_ax, pump_info, windows_per_bin = compute_ssx_axes(
                nxs_in, nbins, ax, (start, stop)
            )

            # Write pump_info
            write_NXnote(nxs_out, "/entry/source/notes", pump_info)

            nxsample = nxentry["sample"]

            # First rotation (attributes already found)
            nxdata.create_dataset(ax, data=rot_ax[ax])
            for key, value in ax_attr.items():
                nxdata[ax].attrs.create(key, value)
            convert_scan_axis(nxsample, nxdata, ax)

            # Check whether multiple windows have been binned together
            if windows_per_bin is not None:
                write_NXnote(
                    nxs_out, "/entry/data", {"windows_per_image": windows_per_bin}
                )
                # And exit here
                return nxs_filename.as_posix()

            # Then translation axes
            for ax_name, ax_range in transl_ax.items():
                nxdata.create_dataset(ax_name, data=ax_range)
                # Get attributes for relevant axis in nxs_in
                ax_attr = dict(nxs_in["/entry/sample/transformations/" + ax_name].attrs)
                for key, value in ax_attr.items():
                    nxdata[ax_name].attrs.create(key, value)
                convert_scan_axis(nxsample, nxdata, ax_name)

    # Run a quick check on axes values and attributes.
    # Some ealy serial Tristan data from March/May 2022 have det_z saved as a string
    with h5py.File(nxs_filename, "r+") as nxs:
        check_and_fix_det_axis(nxs)

    return nxs_filename.as_posix()



from datetime import datetime
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pytest

from nexgen.nxs_write.write_utils import (
    calculate_estimated_end_time,
    calculate_origin,
    create_attributes,
    find_number_of_images,
    mask_and_flatfield_writer,
    mask_and_flatfield_writer_for_event_data,
    set_dependency,
    write_compressed_copy,
)

test_module = {"fast_axis": [1, 0, 0], "slow_axis": [0, 1, 0]}


def test_create_attributes(dummy_nexus_file):
    nxentry = dummy_nexus_file.require_group("/entry/")
    create_attributes(nxentry, ("NX_class", "version"), ("NXentry", "0.0"))

    assert dummy_nexus_file["/entry/"].attrs["NX_class"] == b"NXentry"
    assert dummy_nexus_file["/entry/"].attrs["version"] == b"0.0"


def test_set_dependency(mock_goniometer):
    # Check that the end of the dependency chain always gets set to b"."
    assert set_dependency(".", "/entry/sample/transformations/") == b"."
    assert set_dependency(mock_goniometer.axes_list[0].depends) == b"."
    assert set_dependency("whatever") != b"."

    # Check the path
    assert (
        set_dependency(mock_goniometer.axes_list[-1].depends, "/entry")
        == b"/entry/sam_z"
    )


def test_calculate_origin():
    # Check that the default return offset value is 1.0
    beam_center = [1000, 2000]
    pixel_size = [5e-05, 5e-05]
    vec1, val1 = calculate_origin(
        beam_center, pixel_size, test_module["fast_axis"], test_module["slow_axis"]
    )
    assert len(vec1) == 3
    assert val1 == 1.0


def test_find_number_of_images_returns_0_if_no_file_passed():
    n_images = find_number_of_images([])
    assert n_images == 0


def test_mask_external_link_writer(dummy_nexus_file):
    nxdet_path = "/entry/instrument/detector/"
    nxdetector = dummy_nexus_file.require_group(nxdet_path)
    test_mask = "path/to/mask/filename"
    mask_and_flatfield_writer(nxdetector, "pixel_mask", test_mask, False)

    assert not dummy_nexus_file[nxdet_path + "pixel_mask_applied"][()]


@patch("nexgen.nxs_write.write_utils.write_compressed_copy")
def test_flatfield_dataset_writer(fake_copy_write, dummy_nexus_file):
    nxdet_path = "/entry/instrument/detector/"
    nxdetector = dummy_nexus_file.require_group(nxdet_path)
    test_flatfield = np.ndarray(shape=(3, 3))
    mask_and_flatfield_writer(nxdetector, "flatfield", test_flatfield, True)

    assert dummy_nexus_file[nxdet_path + "flatfield_applied"][()]
    fake_copy_write.assert_called_once()


def test_mask_flatfield_writer_for_events(dummy_nexus_file):
    nxdet_path = "/entry/instrument/detector/"
    nxdetector = dummy_nexus_file.require_group(nxdet_path)

    mask_and_flatfield_writer_for_event_data(
        nxdetector,
        "pixel_mask",
        "path/to/mask.h5",
        False,
        Path(""),
    )

    assert "pixel_mask" in list(dummy_nexus_file[nxdet_path].keys())
    assert "pixel_mask_applied" in list(dummy_nexus_file[nxdet_path].keys())


def test_if_flafield_is_None_nothing_written_for_events(dummy_nexus_file):
    nxdet_path = "/entry/instrument/detector/"
    nxdetector = dummy_nexus_file.require_group(nxdet_path)

    mask_and_flatfield_writer_for_event_data(
        nxdetector,
        "flatfield",
        None,
        False,
        Path(""),
    )

    assert "flatfield" not in list(dummy_nexus_file[nxdet_path].keys())


def test_write_copy_raises_error_if_both_array_and_file():
    with pytest.raises(ValueError):
        write_compressed_copy("", "", np.array([0.0]), "filename")


def test_calculate_estimated_end_time_from_iso_string():
    timestamp_str = "2023-11-15T10:30:42Z"
    est_end_time = calculate_estimated_end_time(timestamp_str, 10)
    assert est_end_time == "2023-11-15T10:30:52Z"


def test_calculate_estimated_end_time_from_datetime():
    timestamp = datetime.strptime("2023-11-15T10:30:42", "%Y-%m-%dT%H:%M:%S")
    est_end_time = calculate_estimated_end_time(timestamp, 20)
    assert est_end_time == "2023-11-15T10:31:02Z"



import tempfile
from unittest.mock import MagicMock

import h5py
import numpy as np
import pytest

from nexgen.tools.VDS_tools import (
    Dataset,
    create_virtual_layout,
    find_datasets_in_file,
    image_vds_writer,
    jungfrau_vds_writer,
    split_datasets,
)


def test_when_get_frames_and_shape_less_than_1000_then_correct():
    sshape = split_datasets(["test1"], (500, 10, 10))
    assert sshape == [Dataset("test1", (500, 10, 10), 0, 500)]

    create_virtual_layout(sshape, np.uint16)


def test_when_get_frames_and_shape_greater_than_1000_then_correct():
    sshape = split_datasets(["test1", "test2"], (1300, 10, 10))
    assert sshape == [
        Dataset("test1", (1000, 10, 10), 0, 1000),
        Dataset("test2", (300, 10, 10), 0, 300),
    ]

    create_virtual_layout(sshape, np.uint16)


def test_when_get_frames_and_shape_less_than_1000_non_zero_then_correct():
    sshape = split_datasets(["test1"], (500, 10, 10), 200, (300, 10, 10))
    assert sshape == [Dataset("test1", (500, 10, 10), 200, 500)]

    create_virtual_layout(sshape, np.uint16)


def test_when_get_frames_and_shape_greater_than_1000_non_zero_then_correct():
    sshape = split_datasets(["test1", "test2"], (1500, 10, 10), 200)
    assert sshape == [
        Dataset("test1", (1000, 10, 10), 200, 1000),
        Dataset("test2", (500, 10, 10), 0, 500),
    ]

    create_virtual_layout(sshape, np.uint16)


def test_when_get_frames_and_shape_greater_than_1000_non_zero_greater_than_1000_then_correct():
    sshape = split_datasets(["test1", "test2"], (1500, 10, 10), 1200)
    assert sshape == [
        Dataset("test2", (500, 10, 10), 200, 500),
    ]

    create_virtual_layout(sshape, np.uint16)


def test_when_get_frames_and_shape_much_greater_than_1000_non_zero_greater_than_1000_then_correct():
    returned = split_datasets(
        ["test1", "test2", "test3", "test4"], (3100, 10, 10), 1100
    )
    assert returned == [
        Dataset("test2", (1000, 10, 10), 100, 1000),
        Dataset("test3", (1000, 10, 10), 0, 1000),
        Dataset("test4", (100, 10, 10), 0, 100),
    ]

    create_virtual_layout(returned, np.uint16)


def test_when_start_idx_higher_than_full_then_exception_raised():
    with pytest.raises(ValueError):
        split_datasets(["test1", "test2"], (1100, 10, 10), 3100)


def test_when_start_idx_negative_then_exception_raised():
    with pytest.raises(ValueError):
        split_datasets(["test1"], (1100, 10, 10), -100)


@pytest.fixture
def nexus_file_with_single_dataset():
    test_hdf_file = tempfile.TemporaryFile()
    test_nexus_file = h5py.File(test_hdf_file, "w")
    test_nexus_file["/entry/data/data_0001"] = h5py.ExternalLink("filename", "path")
    yield test_nexus_file


def test_find_datasets_int_file(nexus_file_with_single_dataset):
    nxdata = nexus_file_with_single_dataset["/entry/data"]
    dsets = find_datasets_in_file(nxdata)
    assert len(dsets) == 1
    assert dsets[0] == "data_0001"


def test_when_float_shape_passed_to_vds_writer_then_no_exception(
    nexus_file_with_single_dataset,
):
    image_vds_writer(nexus_file_with_single_dataset, (500.0, 10.0, 10.0), start_index=0)


def test_when_float_start_index_passed_to_vds_writer_then_no_exception(
    nexus_file_with_single_dataset,
):
    image_vds_writer(
        nexus_file_with_single_dataset, (500.0, 10.0, 10.0), start_index=0.0
    )


def test_given_file_with_no_dataset_external_links_then_exception_is_sensible():
    test_hdf_file = tempfile.TemporaryFile()
    test_nexus_file = h5py.File(test_hdf_file, "w")
    test_nexus_file["/entry/data/data_0001"] = MagicMock()
    with pytest.raises(KeyError):
        image_vds_writer(test_nexus_file, (1000, 10, 10))


def test_jungfrau_vds_writer_with_external_dsets():
    test_hdf_file = tempfile.TemporaryFile()
    test_nexus_file = h5py.File(test_hdf_file, "w")
    test_nexus_file["/entry/data/data_0001"] = MagicMock()
    source_dsets = ["path/to/file1", "path/to/file2"]
    jungfrau_vds_writer(
        test_nexus_file,
        (100, 1066, 1030),
        source_dsets=source_dsets,
    )
    assert "data" in list(test_nexus_file["/entry/data"].keys())


def test_jungfrau_vds_writer_not_failing_if_no_external_dsets():
    test_hdf_file = tempfile.TemporaryFile()
    test_nexus_file = h5py.File(test_hdf_file, "w")
    test_nexus_file["/entry/data/data_0001"] = h5py.ExternalLink("f1le1", "data")
    test_nexus_file["/entry/data/data_0002"] = h5py.ExternalLink("f1le2", "data")
    jungfrau_vds_writer(
        test_nexus_file,
        (100, 1066, 1030),
    )
    assert "data" in list(test_nexus_file["/entry/data"].keys())



"""
General utilities for nexgen
"""

from __future__ import annotations

import re
from collections import namedtuple
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Tuple

import h5py
import numpy as np
import pint
from freephil.common import scope_extract as ScopeExtract  # Define scope extract type
from numpy.typing import ArrayLike

__all__ = [
    "get_filename_template",
    "get_nexus_filename",
    "walk_nxs",
    "units_of_length",
    "units_of_time",
    "get_iso_timestamp",
    "ScopeExtract",
]

MAX_FRAMES_PER_DATASET = 1000
MAX_SUFFIX_DIGITS = 6

# Define coordinates
Point3D = namedtuple("Point3D", ("x", "y", "z"))
Point3D.__doc__ = """Coordinates in 3D space."""

# Filename pattern: filename_######.h5 or filename_meta.h5
# P = re.compile(r"(.*)_(?:\d+)")
P = re.compile(r"(.*)_(?:meta|master|\d+)")


def coerce_to_path(filename: Path | str):
    if not isinstance(filename, Path):
        filename = Path(filename).expanduser().resolve()
    return filename


def find_in_dict(key: str, params_dict: Dict):
    if key in list(params_dict.keys()):
        return True
    return False


def get_filename_template(input_filename: Path) -> str:
    """
    Get the data file name template from either the master or the meta file.

    Args:
        input_filename (Path): Path object containing the name of master or meta file.
                            The format should be either file_master.h5, file.nxs for a master file, file_meta.h5 for a meta file.

    Raises:
        NameError: If the input file does not have the expected format.

    Returns:
        filename_template (str): String template for the name of blank data file.
    """
    if input_filename.suffix == ".nxs":
        filename_root = input_filename.stem
        filename_template = (
            input_filename.parent / f"{filename_root}_%0{MAX_SUFFIX_DIGITS}d.h5"
        )
    elif input_filename.suffix == ".h5" and "master" in input_filename.as_posix():
        filename = input_filename.stem.replace("master", f"%0{MAX_SUFFIX_DIGITS}d")
        filename_template = input_filename.parent / f"{filename}.h5"
    elif input_filename.suffix == ".h5" and "meta" in input_filename.as_posix():
        filename = input_filename.stem.replace("meta", f"%0{MAX_SUFFIX_DIGITS}d")
        filename_template = input_filename.parent / f"{filename}.h5"
    else:
        raise NameError(
            "Input file did not have the expected format for a master or meta file."
        )
    # so that filename_template.as_posix() % 1 will become filename_000001.h5
    return filename_template.as_posix()


def get_nexus_filename(input_filename: Path | str, copy: bool = False) -> Path:
    """
    Get the filename for the NeXus file from the stem of the input file name.

    Args:
        input_filename (Path | str): File name and path of either a .h5 data file or a _meta.h5 file.
        copy (bool, optional): Avoid trying to write a new file with the same name as the old one when making a copy. Defaults to False.

    Returns:
        Path: NeXus file name (.nxs) path.
    """
    input_filename = coerce_to_path(input_filename)
    filename_stem = P.fullmatch(input_filename.stem)
    if filename_stem:
        filename = filename_stem[1]
    else:
        filename = input_filename.stem

    if copy is True:
        nxs_filename = input_filename.parent / f"{filename}_copy.nxs"
    else:
        nxs_filename = input_filename.parent / f"{filename}.nxs"
    return nxs_filename


def walk_nxs(nxs_obj: h5py.File | h5py.Group) -> List[str]:
    """
    Walk all the groups, subgroups and datasets of an object.

    Args:
        nxs_obj (h5py.File | h5py.Group): Object to walk through, could be a file or a group.

    Returns:
        obj_list (List[str]): List of objects found, as strings.
    """
    obj_list = []
    nxs_obj.visit(obj_list.append)
    return obj_list


# Initialize registry and a Quantity constructor
ureg = pint.UnitRegistry()
Q_ = ureg.Quantity


def units_of_length(q: str | float, to_base: bool = False) -> Q_:  # -> pint.Quantity:
    """
    Check that a quantity of length is compatible with NX_LENGTH, defaulting to m if dimensionless.

    Args:
        q (Any): An object that can be interpreted as a pint Quantity, it can be dimensionless.
        to_base (bool, optional): If True, convert to base units of length (m). Defaults to False.

    Raises:
        ValueError: If the input value is a negative number.
        pint.errors.DimensionalityError: If the input value is not a quantity of lenght.

    Returns:
        quantity (pint.Quantity): A pint quantity with units applied if it was dimensionless.
    """
    quantity = Q_(q)
    if quantity <= 0:
        raise ValueError(
            f"Quantity (length) must be positive. Current value: {quantity}."
        )
    quantity = quantity * ureg.m if quantity.dimensionless else quantity
    if quantity.check("[length]"):
        if to_base is True:
            return quantity.to_base_units()
        else:
            return quantity
    else:
        raise pint.errors.DimensionalityError(
            quantity, "a quantity of", quantity.dimensionality, ureg.mm.dimensionality
        )


def units_of_time(q: str) -> Q_:  # -> pint.Quantity:
    """
    Check that a quantity of time is compatible with NX_TIME, defaulting to s if dimensionless.
    Convert to seconds if time is passed as a fraction of it.

    Args:
        q (str): A string that can be interpreted as a pint Quantity, it can be dimensionless.

    Raises:
        ValueError: If the input value is a negative number.
        pint.errors.DimensionalityError: If the input value is not a quantity of lenght.

    Returns:
        quantity (pint.Quantity): A pint quantity in s, with units applied if it was dimensionless.
    """
    quantity = Q_(q)
    if quantity <= 0:
        raise ValueError(
            f"Quantity (time) of time must be positive. Current value: {quantity}."
        )
    quantity = quantity * ureg.s if quantity.dimensionless else quantity
    if quantity.check("[time]"):
        return quantity.to_base_units()
    else:
        raise pint.errors.DimensionalityError(
            quantity, "a quantity of", quantity.dimensionality, ureg.s.dimensionality
        )


def get_iso_timestamp(ts: str | float) -> str:
    """
    Format a timestamp string to be stores in a NeXus file according to ISO8601: 'YY-MM-DDThh:mm:ssZ'

    Args:
        ts (str | float): Input string, can also be a timestamp (eg. time.time()) string.
            Allowed formats: "%Y-%m-%dT%H:%M:%S", "%Y-%m-%d %H:%M:%S", "%a %b %d %Y %H:%M:%S", "%A, %d. %B %Y %I:%M%p".

    Returns:
        ts_iso (str): Formatted timestamp.
    """
    # Format strings for timestamps
    format_list = [
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%d %H:%M:%S",
        "%a %b %d %Y %H:%M:%S",
        "%A, %d. %B %Y %I:%M%p",
    ]
    if ts is None:
        return None
    try:
        ts = float(ts)
        ts_iso = datetime.utcfromtimestamp(ts).replace(microsecond=0).isoformat()
    except ValueError:
        for fmt in format_list:
            try:
                ts_iso = datetime.strptime(ts, fmt).isoformat()
            except ValueError:
                ts_iso = str(ts)
    if ts_iso.endswith("Z") is False:
        ts_iso += "Z"
    return ts_iso


def imgcif2mcstas(vector: List | Tuple | ArrayLike) -> Tuple:
    """
    Convert from the standard coordinate frame used by imgCIF/CBF to the
    NeXus McStas coordinate system.

    Args:
        vector (List | Tuple | np.array): Coordinates to be converted.

    Returns:
        Tuple: Converted coordinate values.
    """
    c2n = np.array([[-1, 0, 0], [0, 1, 0], [0, 0, -1]])
    return tuple(np.dot(c2n, vector))


def coord2mcstas(vector: List | Tuple | ArrayLike, mat: ArrayLike) -> Tuple:
    """
    General conversion from a new coordinate convention to the NeXus McStas coordinate system.

    Args:
        vector (List | Tuple | np.array): Coordinates to be converted.
        mat (np.ndarray): Coordinate transformation matrix.

    Returns:
        Tuple: Converted coordinate values
    """
    return tuple(np.dot(mat, vector))



"""
Command line tool to generate NXmx NeXus files for Serial Crystallography.
"""

import argparse
import logging
from pathlib import Path
from typing import Tuple

from .. import log
from ..beamlines.SSX_chip import CHIP_DICT_DEFAULT
from ..utils import P
from . import version_parser

logger = logging.getLogger("nexgen.SSX_cli")

usage = "%(prog)s <sub-command> collection-data [options]"
parser = argparse.ArgumentParser(
    usage=usage, description=__doc__, parents=[version_parser]
)
parser.add_argument("--debug", action="store_const", const=True)


def eiger_collection(args):
    logger.info("Create a NeXus file for SSX collection on Eiger.")
    from ..beamlines.SSX_Eiger_nxs import ssx_eiger_writer

    ssx_eiger_writer(
        Path(args.visitpath).expanduser().resolve(),
        args.filename_root,
        args.beamline,
        expt_type=args.expt_type,
        pump_status=args.pump_status,
        num_imgs=args.num_imgs,
        exp_time=args.exp_time,
        det_dist=args.det_dist,
        beam_center=args.beam_center,
        transmission=args.transmission,
        wavelength=args.wavelength,
        start_time=args.start,
        stop_time=args.stop,
        chip_info=CHIP_DICT_DEFAULT,  # TODO This might be better passed as a json/yaml or whatever
        chipmap=args.chipmap,
    )


def tristan_collection(args):
    logger.info("Create a NeXus file for SSX collection on Tristan.")
    from ..beamlines.SSX_Tristan_nxs import ssx_tristan_writer

    ssx_tristan_writer(
        Path(args.visitpath).expanduser().resolve(),
        args.filename_root,
        args.beamline,
        xp_time=args.exp_time,
        det_dist=args.det_dist,
        beam_center=args.beam_center,
        transmission=args.transmission,
        wavelength=args.wavelength,
        start_time=args.start,
        stop_time=args.stop,
        chip_info=CHIP_DICT_DEFAULT,  # TODO This might be better passed as a json/yaml or whatever
        chipmap=args.chipmap,
    )


# Define a parser for the basic collection parameters
class _ImportCollect(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        input_file, visitpath, filename_root = self.find_import_args(values)
        setattr(namespace, self.dest, input_file)
        setattr(namespace, "visitpath", visitpath)
        setattr(namespace, "filename_root", filename_root)

    @staticmethod
    def find_import_args(val) -> Tuple[str]:
        input_file = Path(val).expanduser().resolve()
        visitpath = input_file.parent
        filename_root = P.fullmatch(input_file.stem)[1]
        return input_file, visitpath, filename_root


collect_parser = argparse.ArgumentParser(add_help=False)
# Get visitpath and filename_root out of meta file
collect_parser.add_argument(
    "input_file",
    type=str,
    action=_ImportCollect,
    help="Path to _meta.h5 or _000001.h5 file.",
)
collect_parser.add_argument("beamline", type=str, help="Beamline name.")
collect_parser.add_argument(
    "-det",
    "--det-dist",
    type=float,
    help="Detector distance in mm.",
)
collect_parser.add_argument(
    "-tr",
    "--transmission",
    type=float,
    default=None,
    help="Attenuator transmission.",
)
collect_parser.add_argument(
    "-wl",
    "--wavelength",
    type=float,
    default=None,
    help="Incident beam wavelength, in A.",
)
collect_parser.add_argument(
    "-bc",
    "--beam-center",
    type=float,
    nargs=2,
    default=[0.0, 0.0],
    help="Beam center (x,y) positions.",
)
collect_parser.add_argument(
    "--start", "--start-time", type=str, default=None, help="Collection start time."
)
collect_parser.add_argument(
    "--stop", "--stop-time", type=str, default=None, help="Collection end time."
)

# Define subparsers
subparsers = parser.add_subparsers(
    help="Choose the writer based on the Detector in use on the beamline. \
        Run SSX_nexus <command> --help to see the parameters for each sub-command.",
    required=True,
    dest="sub-command",
)

eiger_parser = subparsers.add_parser(
    "1",
    aliases=["eiger"],
    description=("Trigger Eiger writer."),
    parents=[collect_parser],
)
eiger_parser.add_argument(
    "expt_type",
    type=str,
    choices=["extruder", "fixed-target", "3Dgridscan"],
    help="Choose the type of serial experiment.",
)
eiger_parser.add_argument("num_imgs", type=int, help="Total number of images.")
eiger_parser.add_argument(
    "-e",
    "--exp-time",
    type=float,
    required=True,
    help="Exposure time in s.",
)
eiger_parser.add_argument(
    "-p",
    "--pump-status",
    action="store_true",
    default=False,
    help="Select pump status.",
)
eiger_parser.add_argument("--chipmap", type=str, help="Location of chipmap.")
eiger_parser.set_defaults(func=eiger_collection)

tristan_parser = subparsers.add_parser(
    "2",
    aliases=["tristan"],
    description=("Trigger Tristan writer."),
    parents=[collect_parser],
)
tristan_parser.add_argument(
    "-e",
    "--exp-time",
    type=float,
    required=True,
    help="Total collection time in s.",
)
tristan_parser.add_argument("--chipmap", type=str, help="Location of chipmap.")
tristan_parser.set_defaults(func=tristan_collection)


def main():
    log.config()
    args = parser.parse_args()
    args.func(args)



from datetime import datetime
from unittest.mock import patch

import numpy as np

from nexgen.nxs_utils import Axis, Goniometer, TransformationType

fake_gonio = Goniometer(
    [Axis("omega", ".", TransformationType.ROTATION, (0, 0, -1), 0.0)],
    {"omega": np.arange(0, 10, 1)},
)


@patch("nexgen.nxs_write.NXmxWriter.write_NXsample")
@patch("nexgen.nxs_write.NXmxWriter.write_NXsource")
@patch("nexgen.nxs_write.NXmxWriter.write_NXdetector_module")
@patch("nexgen.nxs_write.NXmxWriter.write_NXdetector")
@patch("nexgen.nxs_write.NXmxWriter.write_NXinstrument")
@patch("nexgen.nxs_write.NXmxWriter.write_NXdata")
@patch("nexgen.nxs_write.NXmxWriter.write_NXentry")
def test_NXmxFileWriter_write(
    mock_NXentry,
    mock_NXdata,
    mock_NXinstrument,
    mock_NXdetector,
    mock_NXmodule,
    mock_NXsource,
    mock_NXsample,
    dummy_NXmxWriter,
):
    dummy_NXmxWriter.goniometer = fake_gonio
    dummy_NXmxWriter.write(write_mode="w")
    mock_NXentry.assert_called_once()
    mock_NXdata.assert_called_once()
    mock_NXinstrument.assert_called_once()
    mock_NXdetector.assert_called_once()
    mock_NXmodule.assert_called_once()
    mock_NXsource.assert_called_once()
    mock_NXsample.assert_called_once()


@patch("nexgen.nxs_write.NXmxWriter.write_NXdatetime")
def test_NXmxFileWriter_updates_timestamps(mock_NXdatetime, dummy_NXmxWriter):
    fake_timestap = datetime.now()
    dummy_NXmxWriter.update_timestamps(fake_timestap, "start_time")
    mock_NXdatetime.assert_called_once()


@patch("nexgen.nxs_write.NXmxWriter.write_NXnote")
def test_NXmxFileWriter_adds_note(mock_NXnote, dummy_NXmxWriter):
    fake_notes = {"foo": "bar"}
    dummy_NXmxWriter.add_NXnote(fake_notes)
    mock_NXnote.assert_called_once()


@patch("nexgen.nxs_write.NXmxWriter.image_vds_writer")
def test_NXmxFileWriter_write_vds(mock_vds_writer, dummy_NXmxWriter):
    dummy_NXmxWriter.goniometer = fake_gonio
    dummy_NXmxWriter.detector.detector_params.image_size = (100, 100)
    dummy_NXmxWriter.write_vds()
    mock_vds_writer.assert_called_once()



"""
Tools to read a chip and compute the coordinates of a Serial Crystallography collection.
"""
from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Tuple

from ..nxs_utils.ScanUtils import ScanDirection

# I24 chip tools

CHIP_DICT_DEFAULT = {
    "X_NUM_STEPS": [0, 20],
    "Y_NUM_STEPS": [0, 20],
    "X_STEP_SIZE": [0, 0.125],
    "Y_STEP_SIZE": [0, 0.125],
    "X_START": [0, 0],
    "Y_START": [0, 0],
    "Z_START": [0, 0],
    "X_NUM_BLOCKS": [0, 8],
    "Y_NUM_BLOCKS": [0, 8],
    "X_BLOCK_SIZE": [0, 3.175],
    "Y_BLOCK_SIZE": [0, 3.175],
    "N_EXPOSURES": [0, 1],
    "PUMP_REPEAT": [0, 0],
}


@dataclass
class Chip:
    """
    Define a fixed target chip.

    Args:
        name (str): Description of the chip.
        num_steps (List[int] | Tuple[int]): Number of windows in each block.
        step_size (List[float] | Tuple[float]): Size of each window (distance between the centers in x and y direction).
        num_blocks (List[int] | Tuple[int]): Total number of blocks in the chip.
        block_size (List[int] | Tuple[int]): Size of each block.
        start_pos (List[float]): Start coordinates (x,y,z)
    """

    name: str

    num_steps: List[int, int] | Tuple[int, int]
    step_size: List[float, float] | Tuple[float, float]
    num_blocks: List[int, int] | Tuple[int, int]
    block_size: List[float, float] | Tuple[float, float]

    start_pos: List[float, float, float] = field(default_factory=[0.0, 0.0, 0.0])

    def tot_blocks(self) -> int:
        return self.num_blocks[0] * self.num_blocks[1]

    def tot_windows_per_block(self) -> int:
        return self.num_steps[0] * self.num_steps[1]

    def window_size(self) -> Tuple[float]:
        return (
            self.num_steps[0] * self.step_size[0],
            self.num_steps[1] * self.step_size[1],
        )

    def chip_size(self) -> Tuple[float]:
        return (
            self.num_blocks[0] * self.block_size[0],
            self.num_blocks[1] * self.block_size[1],
        )


def fullchip_conversion_table(chip: Chip) -> Dict:
    """Associate block coordinates to block number for a full chip.

    Args:
        chip (Chip): General description of the chip.

    Returns:
        Dict: Conversion table, keys are block numbers, values are coordinates.
    """
    coords = []
    table = {}
    for i in range(chip.num_blocks[0]):
        if i % 2 == 0:
            for j in range(chip.num_blocks[1]):
                coords.append((i, j))
        else:
            for j in range(chip.num_blocks[1] - 1, -1, -1):
                coords.append((i, j))
    for k, v in zip(range(1, chip.tot_blocks() + 1), coords):
        table[f"%0{2}d" % k] = v
    return table


def read_chip_map(mapfile: Path | str, x_blocks: int, y_blocks: int) -> Dict:
    """
    Read the .map file for the current collection on a chip.

    Args:
        mapfile (Path | str): Path to .map file. If None, assumes fullchip.
        x_blocks (int): Total number of blocks in x direction in the chip.
        y_blocks (int): Total number of blocks in y direction in the chip.

    Returns:
        Dict: A dictionary whose values indicate either the coordinates on the chip \
            of the scanned blocks, or a string indicating that the whole chip is being scanned.
    """
    if mapfile is None:
        # Assume it's a full chip
        return {"all": "fullchip"}

    with open(mapfile) as f:
        chipmap = f.read()

    block_list = []
    max_num_blocks = x_blocks * y_blocks
    for n, line in enumerate(chipmap.rsplit("\n")):
        if n == max_num_blocks:
            break
        k = line[:2]
        v = line[-1:]
        if v == "1":
            block_list.append(k)
    if len(block_list) == max_num_blocks:
        # blocks["fullchip"] = len(block_list)
        return {"all": "fullchip"}

    blocks = {}
    for b in block_list:
        x = int(b) // x_blocks if int(b) % x_blocks != 0 else (int(b) // x_blocks) - 1
        if x % 2 == 0:
            val = x * x_blocks + 1
            y = int(b) - val
        else:
            val = (x + 1) * x_blocks
            y = val - int(b)
        blocks[b] = (x, y)
    return blocks


def fullchip_blocks_conversion(blocks: Dict[Tuple, Any], chip: Chip) -> Dict:
    new_blocks = {}
    table = fullchip_conversion_table(chip)
    for kt, vt in table.items():
        for kb, vb in blocks.items():
            if kb == vt:
                new_blocks[kt] = vb
    return new_blocks


def compute_goniometer(
    chip: Chip,
    blocks: Dict | None = None,
    full: bool = False,
    ax1: str = "sam_y",
    ax2: str = "sam_x",
) -> Dict[Dict[str | Tuple, float | int]]:
    """Compute the start coordinates of each block in a chip scan.

    The function returns a dictionary associating a list of axes start values \
    and a scan direction to each scanned block.
    If full is True, the blocks argument will be overridden and coordinates will be \
    calculated for every block in the chip.

    Args:
        chip (Chip): General description of the chip schematics: number and size of blocks, \
            size and step of each window, start positions.
        blocks (Dict | None, optional): Scanned blocks. Defaults to None.
        full (bool, optional): True if all blocks have been scanned. Defaults to False.
        ax1 (str, optional): Axis name corrsponding to slow varying axis. Defaults to "sam_y".
        ax2 (str, optional): Axis name corrsponding to fast varying axis. Defaults to "sam_x".

    Returns:
        Dict[Dict[str | Tuple, float | int]]: Axes start coordinates and scan direction of each block. \
            eg. \
                {
                    '01'/(0,0): {
                        'ax1': 0.0,
                        'ax2': 0.0,
                        'direction': 1,
                    }
                }
    """
    x0 = chip.start_pos[0]
    y0 = chip.start_pos[1]

    axes_starts = {}

    if full is True:
        for x in range(chip.num_blocks[0]):
            x_start = x0 + x * chip.block_size[0]
            if x % 2 == 0:
                for y in range(chip.num_blocks[1]):
                    sd = ScanDirection.POSITIVE
                    y_start = y0 + y * chip.block_size[1]
                    axes_starts[(x, y)] = {
                        ax1: round(y_start, 3),
                        ax2: round(x_start, 3),
                        "direction": sd.value,
                    }
            else:
                for y in range(chip.num_blocks[1] - 1, -1, -1):
                    sd = ScanDirection.NEGATIVE
                    y_end = y0 + y * chip.block_size[1]
                    y_start = y_end + (chip.num_steps[1] - 1) * chip.step_size[1]
                    axes_starts[(x, y)] = {
                        ax1: round(y_start, 3),
                        ax2: round(x_start, 3),
                        "direction": sd.value,
                    }
    else:
        for k, v in blocks.items():
            x_start = x0 + v[0] * chip.block_size[0]
            if v[0] % 2 == 0:
                sd = ScanDirection.POSITIVE
                y_start = y0 + v[1] * chip.block_size[1]
            else:
                sd = ScanDirection.NEGATIVE
                y_end = y0 + v[1] * chip.block_size[1]
                y_start = y_end + (chip.num_steps[1] - 1) * (chip.step_size[1])
            axes_starts[k] = {
                ax1: round(y_start, 3),
                ax2: round(x_start, 3),
                "direction": sd.value,
            }

    return axes_starts
